
#%%
#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import xarray as xr
import numpy as np
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
import pickle
from datetime import timedelta, datetime
from dateutil.relativedelta import relativedelta
# DEV: override installed teslakit
import sys
sys.path.insert(0, op.join(os.path.abspath(''), '..', '..'))

# teslakit
from teslakit.database import Database
from teslakit.rbf import RBF_Interpolation
from teslakit.waves import Intradaily_Hydrograph
from teslakit.io.matlab import ReadMatfile
from teslakit.util.time_operations import date2datenum as d2d


## Database and Site parameters
# %%
# --------------------------------------
# Teslakit database

p_data = r'/Users/anacrueda/Documents/Proyectos/TESLA/Teslakit/data'
db = Database(p_data)

# set site
db.SetSite('KWAJALEIN')

# %%
# --------------------------------------
# load data and set parameters
#  wave data reconstructed a 30m water depth

p_sim = r'/Users/anacrueda/Documents/Proyectos/SERDP/nearshores_waves_albaR/'

p_dataset = op.join(p_sim, 'sim10_propagg21.pkl')
fileObject = open(p_dataset,'rb')
subset = pickle.load(fileObject)

p_datasetO = op.join(p_sim, 'complete_h_offshore.nc')
dataO = xr.open_dataset(p_datasetO)

p_dataO_storm = op.join(p_sim, 'SIM_WAVES_TCs_offshore.nc')
data_storm = xr.open_dataset(p_dataO_storm)

#p_tcs = op.join(p_sim, 'SIM_TCs.nc')
#data_tcs = xr.open_dataset(p_tcs)

# %% load slr data

p_slr = r'/Users/anacrueda/Documents/Proyectos/SERDP/niveles_ref_proyecc/slr_kwa.pkl'
file_slr = open(p_slr,'rb')
data_slr = pickle.load(file_slr)

data_slr = data_slr.sel(scenario='0.5')

print(data_slr)

# %% modify time in simulations to be able to compare them

fechas_nuevas = [datetime(1700, 1, 1, 00) + timedelta(hours=x) for x in range(int(len(dataO.time)))]

dataO['time'] = fechas_nuevas


# %% varying AT, SS and MMSL for each simulation

Level_xds = xr.Dataset({})
for i_sim in dataO.n_sim:
    dataO_storm = dataO.sel(n_sim=i_sim, time=data_storm.time, drop=True)
    data_storm_time_h_rand = [

        d2d(d) + timedelta(hours=np.random.randint(23)) for d in data_storm.time.values[:]

    ]
    # el ultimo valor no puede ser random porque la serie sintetica acaba ese dia a las 00
    # OJO chapuza! nico help!!
    data_storm_time_h_rand[-1] = datetime(2700, 1, 1, 00)
    AT = dataO.sel(n_sim=i_sim, time=data_storm_time_h_rand[:]).AT

    Level_s = dataO_storm.SS + dataO_storm.MMSL + AT

    Level = xr.Dataset({
        'level': ('time', xr.DataArray(Level_s)),
    },
        coords={'time': data_storm.time}
    )
    #print(Level)
    #print(i_sim)
    Level_xds = xr.concat([Level_xds, Level], dim='n_sim')

# %% get same time in simulations and slr scenarios

HS = np.reshape([subset['Hs']], (10,len(data_storm.time)),'C')
TP = np.reshape([subset['Tp']], (10,len(data_storm.time)),'C')
DIR = np.reshape([subset['Dir']], (10,len(data_storm.time)),'C')
#LEVEL = np.reshape([Level_xds['level']], (10,len(data_storm.time)),'C')

wvs_agr = xr.Dataset(
    {'Hs': (('n_sim', 'time'), HS),
     'Tp': (('n_sim', 'time'), TP),
     'Dir': (('n_sim', 'time'), DIR),
     'Level':(('n_sim', 'time'), Level_xds['level'])
},
    coords={'n_sim': dataO.n_sim, 'time': data_storm.time}
)


# %% For each simulation modify the dates to fit the slr projections

wvs_agr_slr = wvs_agr.sel(n_sim=0)
fechas_eventos = wvs_agr_slr.time.sel(time=slice('2600', '2699'))
data_storm_time_slr = [

        d2d(d)- relativedelta(years=601) for d in fechas_eventos.time.values[:]

    ]

# %%
wvs_cent =  xr.Dataset({})
for i_sim in range(0,10):
    if i_sim == 0:
        wvs = wvs_agr_slr.sel(time=slice('1700', '1799'))
        # data_storm_time_slr = [
        #
        #     d2d(d) + relativedelta(years=301) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 1:
        wvs = wvs_agr_slr.sel(time=slice('1800', '1899'))
        # data_storm_time_slr = [
        #
        #     d2d(d) + relativedelta(years=201) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 2:
        wvs = wvs_agr_slr.sel(time=slice('1900', '1999'))
        # data_storm_time_slr = [
        #
        #     d2d(d) + relativedelta(years=101) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 3:
        wvs = wvs_agr_slr.sel(time=slice('2000', '2099'))
        # data_storm_time_slr = [
        #
        #     d2d(d) + relativedelta(years=1) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 4:
        wvs = wvs_agr_slr.sel(time=slice('2100', '2199'))
        # data_storm_time_slr = [
        #
        #     d2d(d) - relativedelta(years=101) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 5:
        wvs = wvs_agr_slr.sel(time=slice('2200', '2299'))
        # data_storm_time_slr = [
        #
        #     d2d(d) - relativedelta(years=201) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 6:
        wvs = wvs_agr_slr.sel(time=slice('2300', '2399'))
        # data_storm_time_slr = [
        #
        #     d2d(d) - relativedelta(years=301) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 7:
        wvs = wvs_agr_slr.sel(time=slice('2400', '2499'))
        # data_storm_time_slr = [
        #
        #     d2d(d) - relativedelta(years=401) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 8:
        wvs = wvs_agr_slr.sel(time=slice('2500', '2599'))
        # data_storm_time_slr = [
        #
        #     d2d(d) - relativedelta(years=501) for d in wvs.time.values[:]
        #
        # ]
    elif i_sim == 9:
        wvs = wvs_agr_slr.sel(time=slice('2600', '2699'))
        # data_storm_time_slr = [
        #
        #     d2d(d) - relativedelta(years=601) for d in wvs.time.values[:]
        #
        # ]


    wvs_temp = xr.Dataset(
    {'Hs': ('time', wvs['Hs'][0:len(data_storm_time_slr)]),
     'Tp': ('time', wvs['Tp'][0:len(data_storm_time_slr)]),
     'Dir': ('time', wvs['Dir'][0:len(data_storm_time_slr)]),
     'Level': ('time', wvs['Level'][0:len(data_storm_time_slr)]),
     #'time': ('time', data_storm_time_slr)
    },
    coords={'time': data_storm_time_slr}
    )

    wvs_cent = xr.concat([wvs_cent, wvs_temp], dim='n_sim')


# %%

_, index = np.unique(wvs_cent['time'], return_index=True)

print(index)
print(len(index))
wvs_cent = wvs_cent.isel(time=index)


# %%


merged_xds = xr.merge([wvs_cent, data_slr], compat='no_conflicts', join='inner')

print(merged_xds)


# %%
#define reef characteristics
Rslope = 0.0505*np.ones(len(merged_xds.time))
Bslope = 0.1667*np.ones(len(merged_xds.time))
Rw = 250*np.ones(len(merged_xds.time))
Cf = 0.0105*np.ones(len(merged_xds.time))

print(len(merged_xds.time))

# %%
merged_xds['Level_SLR'] = merged_xds.Level + merged_xds.slr

hs_lo2 = merged_xds['Hs']/(1.5613*merged_xds['Tp']**2)

xds_subset = xr.Dataset(
    {'hs':(('n_sim', 'time'), merged_xds['Hs']),
     'tp':(('n_sim', 'time'), merged_xds['Tp']),
     'dir':(('n_sim', 'time'), merged_xds['Dir']),
     'level':(('n_sim', 'time'), merged_xds['Level_SLR']),
     'hs_lo2': hs_lo2,
     'rslope': Rslope,
     'bslope': Bslope,
     'rwidth': Rw,
     'cf': Cf,
},
    coords={'n_sim': merged_xds.n_sim, 'time': merged_xds.time}
)

print(xds_subset)

# %%

# load min, max, of simulated reef and wave conditions and rbfs coefficients

p_rbf = r'/Users/anacrueda/Documents/Proyectos/SERDP/HyCreWWcode/RBF_coefficients/'

p_min = op.join(p_rbf, 'Min_from_simulations.mat')
p_max = op.join(p_rbf, 'Max_from_simulations.mat')

smin = ReadMatfile(p_min)
smax = ReadMatfile(p_max)

# data max and min
dl = np.column_stack([smin['minimum'], smax['maximum']])

print(dl)

d = {
    'hs': dl[1],
    'tp': dl[2],
    'level':dl[0],
    'rslope': dl[3],
    'bslope': dl[4],
    'rwidth': dl[5],
    'cf': dl[6],
    'hs_lo2': [0.005, 0.05],
}

for k in d.keys():
    print(k, d[k])

# %%
# input data quality check
#['level','hs','tp','rslope','bslope','rwidth','cf']:

runup_all_xds = xr.Dataset({})

for i_sim in xds_subset.n_sim:
    xds_f = xds_subset.sel(n_sim=i_sim)

    for vn in xds_f.variables:
        if vn in d.keys():
            # get limits from dictionary
            l, h = tuple(d[vn])

            # find positions inside limits
            xds_f[vn] = xds_f[vn].where((xds_f[vn] >= l) & (xds_f[vn] <= h))

    print('xds_f')
    print(xds_f)
    print()
    # Normalize reef and water level parameters data
    xds_reef = xr.Dataset({
       'level': (xds_f['level']),
       'rslope': (xds_f['rslope']),
       'bslope': (xds_f['bslope']),
       'rwidth': (xds_f['rwidth']),
       'cf': (xds_f['cf']),
        },
         coords={'time': xds_f['time']}
        )

    print('xds_reef')
    print(xds_reef)
    print()
    xds_n = xds_reef.copy()
    for k in xds_reef.variables:
        if k in d.keys():
           v = xds_n[k].values[:]
           l, h = tuple(d[k])
           xds_n[k] =(('time',), (v-l)/(h-l))
    print('xds_n')
    print(xds_n)

    # Hycreeww
    # 15 different wave conditions were considered for each water level and reef characteristics
    # the runup is calculated and then it is linearly interpolated for the actual wave conditions.
    ncases = 15
    runup = []
    # Build RBF target numpy array
    target = np.column_stack(
         (xds_n['level'], xds_n['rslope'], xds_n['bslope'], xds_n['rwidth'], xds_n['cf'])
    )
    target_wavecon = np.column_stack(
      (xds_f['hs'],xds_f['tp'],xds_f['hs_lo2'])
    )

    f_tot = []
    for j in range(0, ncases):
        p_coeff= op.join(p_rbf, 'Coeffs_Runup_Xbeach_test' + str(j+1) + '.mat')
        coeff = ReadMatfile(p_coeff)
        rbf_constant = coeff['coeffsave']['RBFConstant']
        rbf_coeff = coeff['coeffsave']['rbfcoeff']
        nodes = coeff['coeffsave']['x']
        x = target
        f_temp = RBF_Interpolation(rbf_constant, rbf_coeff, nodes,x.T)
        f_tot.append(f_temp)
    print(f_tot)

    #linear interpolation
    hs=[1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5] #Wave heigth conditions used in RBF
    hs_lo=[0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05, 0.005, 0.025, 0.05 ] #Wave conditions used in RBF

    #print(np.shape(f_tot))
    #print(target_wavecon.shape)
    to_z = np.array(f_tot)
    #print(to_z.shape)
    RU = []
    for j in range (0,len(target)):
        x=hs; y=hs_lo; z=to_z[:,j]
        hs_e=target_wavecon[j,0]
        #tp_e=target_wavecon[j,1]
        hs_lo_e=target_wavecon[j,2]
        vq = griddata((x,y),z,(hs_e,hs_lo_e),method='linear')
        RU.append(vq)
    print(RU)
    print()
    #print(i_sim)

    runup_x = np.array(RU)

    runup_xds = xr.Dataset({
        'runup': ('time', xr.DataArray(runup_x)),
    },
        coords={'time': xds_subset.time}
    )
    runup_all_xds = xr.concat([runup_all_xds, runup_xds], dim='n_sim')


# %%

true_runup = runup_all_xds['runup'] + xds_subset['level']

# we need to add the level to the hycreww estimation of Ru to get the proper runup
xds_results = xr.Dataset({
        'hs': (('n_sim','time'), xds_subset['hs']),
        'tp': (('n_sim','time'), xds_subset['tp']),
        'dir': (('n_sim','time'), xds_subset['dir']),
        'level': (('n_sim','time'), xds_subset['level']),
        'hs_lo2': (('n_sim','time'), xds_subset['hs_lo2']),
        'rslope': (xds_subset['rslope']),
        'bslope': (xds_subset['bslope']),
        'rwidth': (xds_subset['rwidth']),
        'cf': (xds_subset['cf']),
        'runup': (('n_sim','time'), true_runup),
        },

         coords={'n_sim': xds_subset.n_sim, 'time': xds_subset.time},

    )

print(xds_results)

# %%



fig, ax = plt.subplots(1,1, figsize=(12, 9))
ax.plot(xds_results.time, xds_results.runup[9,:], '.')
#ax.plot(time_axis_H, y_fit_H, label=scenario)

ax.set_xlim([time_axis[0], time_axis[-3]])

ax.set_ylim([0, 4])
ax.set_title(site + ', runup with SL rise', fontweight='bold')
ax.set_ylabel('m')
ax.grid()
plt.show()
#fig.savefig(os.path.join(rutin, 'SLR_' + site + '.png'))


# %%



fig, ax = plt.subplots(1,1, figsize=(12, 9))
ax.plot(merged_xds.time, merged_xds.slr, '-')
ax.plot(merged_xds.time, merged_xds.Level_SLR[5,:],'.r')
plt.show()

# %%


# ANNUAL MAXIMA

sim_A = xds_results['runup'].groupby('time.year').max(dim='time')

def axplot_AM_SLR(ax, t_s, v_s, var_name):
    'axes plot runup annual maxima for a SLR scenario'

    # simulation maxima - mean
    mn = np.mean(v_s, axis=0)
    ax.plot(
        t_s, mn, '-r',
        linewidth = 2, label = 'Simulation (mean)',
        zorder=8,
    )

    # simulation maxima percentile 95% and 05%
    p95 = np.percentile(v_s, 95, axis=0,)
    p05 = np.percentile(v_s, 5, axis=0,)

    ax.plot(
        t_s, p95, linestyle='-', color='grey',
        linewidth = 2, #label = 'Simulation (95% percentile)',
    )

    ax.plot(
        t_s, p05, linestyle='-', color='grey',
        linewidth = 2, # label = 'Simulation (05% percentile)',
    )
    ax.fill_between(
        t_s, p05, p95, color='lightgray',
        label = 'Simulation (05 - 95 percentile)'
    )

    # customize axs
    ax.legend(loc='lower right')
    ax.set_title('Annual Maxima', fontweight='bold')
    ax.set_xlabel('Time (years)')
    ax.set_ylabel(' runup (m)')
    #ax.set_xlim(left=t_s[0], right=t_s[-1])))
    ax.tick_params(axis='both', which='both', top=True, right=True)
    ax.grid(which='both')


t_s = sim_A['year']
v_s = sim_A


# figure

# common fig parameters
_faspect = 1.618
_fsize = 9.8
_fdpi = 128
name = 'runup'

fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

axplot_AM_SLR(
        axs,
        t_s, v_s,
        name,
)
plt.show()


import os.path as op
import tempfile
import unittest
from pathlib import Path

import pandas as pd

from bluemath_tk.downloaders.noaa.noaa_downloader import NOAADownloader


class TestNOAADownloader(unittest.TestCase):
    def setUp(self):
        """Set up test fixtures before each test method."""
        self.temp_dir = tempfile.mkdtemp()
        self.downloader = NOAADownloader(
            base_path_to_download=self.temp_dir,
            debug=True,
            check=False,  # Just check paths to download
        )

    def test_download_bulk_parameters(self):
        """Test downloading bulk parameters."""

        # Test without loading DataFrame
        result = self.downloader.download_data(
            data_type="bulk_parameters",
            buoy_id="41001",
            years=[2023],
        )
        self.assertIsNotNone(result)
        self.assertIsInstance(result, str)
        print(f"\nBulk parameters download result: {result}")

        # Test with loading DataFrame
        df = self.downloader.download_data(
            data_type="bulk_parameters",
            buoy_id="41001",
            years=[2023],
            load_df=True,
        )
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertTrue("datetime" in df.columns)
        self.assertTrue(len(df) > 0)
        print(f"\nBulk parameters DataFrame shape: {df.shape}")

    def test_download_wave_spectra(self):
        """Test downloading wave spectra."""

        # Test without loading DataFrame
        result = self.downloader.download_data(
            data_type="wave_spectra",
            buoy_id="41001",
            years=[2023],
        )
        self.assertIsNotNone(result)
        self.assertIsInstance(result, str)
        print(f"\nWave spectra download result: {result}")

        # Test with loading DataFrame
        df = self.downloader.download_data(
            data_type="wave_spectra",
            buoy_id="41001",
            years=[2023],
            load_df=True,
        )
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertTrue(isinstance(df.index, pd.DatetimeIndex))
        self.assertTrue(len(df) > 0)
        print(f"\nWave spectra DataFrame shape: {df.shape}")

    def test_download_directional_spectra(self):
        """Test downloading directional spectra."""

        # Test without loading DataFrame
        result = self.downloader.download_data(
            data_type="directional_spectra",
            buoy_id="41001",
            years=[2023],
        )
        self.assertIsNotNone(result)
        self.assertIsInstance(result, str)
        print(f"\nDirectional spectra download result: {result}")

        # Test with loading DataFrame
        alpha1, alpha2, r1, r2, c11 = self.downloader.download_data(
            data_type="directional_spectra",
            buoy_id="41001",
            years=[2023],
            load_df=True,
        )
        # Check each coefficient DataFrame
        for name, df in [
            ("alpha1", alpha1),
            ("alpha2", alpha2),
            ("r1", r1),
            ("r2", r2),
            ("c11", c11),
        ]:
            if df is not None:
                self.assertIsInstance(df, pd.DataFrame)
                self.assertTrue(isinstance(df.index, pd.DatetimeIndex))
                self.assertTrue(len(df) > 0)
                print(f"\n{name} DataFrame shape: {df.shape}")

    def test_multiple_years_loading(self):
        """Test loading multiple years of data."""

        # Test bulk parameters with multiple years
        df = self.downloader.download_data(
            data_type="bulk_parameters",
            buoy_id="41001",
            years=[2022, 2023],
            load_df=True,
        )
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertTrue("datetime" in df.columns)
        self.assertTrue(len(df) > 0)

        # Check that data spans multiple years
        years = df["datetime"].dt.year.unique()
        self.assertTrue(len(years) > 1)
        print(f"\nBulk parameters multiple years: {sorted(years)}")

        # Test wave spectra with multiple years
        df = self.downloader.download_data(
            data_type="wave_spectra",
            buoy_id="41001",
            years=[2022, 2023],
            load_df=True,
        )
        self.assertIsNotNone(df)
        self.assertIsInstance(df, pd.DataFrame)
        self.assertTrue(isinstance(df.index, pd.DatetimeIndex))
        self.assertTrue(len(df) > 0)

        # Check that data spans multiple years
        years = df.index.year.unique()
        self.assertTrue(len(years) > 1)
        print(f"\nWave spectra multiple years: {sorted(years)}")

    def test_list_data_types(self):
        """Test listing available data types."""

        data_types = self.downloader.list_data_types()
        self.assertIsInstance(data_types, list)
        self.assertTrue(len(data_types) > 0)
        print(f"\nAvailable data types: {data_types}")

    def test_list_datasets(self):
        """Test listing available datasets."""

        datasets = self.downloader.list_datasets()
        self.assertIsInstance(datasets, list)
        self.assertTrue(len(datasets) > 0)
        print(f"\nAvailable datasets: {datasets}")

    def test_show_markdown_table(self):
        """Test showing markdown table."""

        self.downloader.show_markdown_table()

    def test_file_paths(self):
        """Test that downloaded files exist in the correct locations."""

        # Download data
        self.downloader.download_data(
            data_type="bulk_parameters",
            buoy_id="41001",
            years=[2023],
        )

        # Check bulk parameters file
        bulk_file = op.join(
            self.temp_dir,
            "buoy_data",
            "41001",
            "buoy_41001_bulk_parameters.csv",
        )
        self.assertTrue(op.exists(bulk_file))
        print(f"\nBulk parameters file exists: {bulk_file}")

        # Download and check wave spectra
        self.downloader.download_data(
            data_type="wave_spectra",
            buoy_id="41001",
            years=[2023],
        )
        wave_file = op.join(
            self.temp_dir,
            "buoy_data",
            "41001",
            "wave_spectra",
            "buoy_41001_spectra_2023.csv",
        )
        self.assertTrue(op.exists(wave_file))
        print(f"\nWave spectra file exists: {wave_file}")

        # Download and check directional spectra
        self.downloader.download_data(
            data_type="directional_spectra",
            buoy_id="41001",
            years=[2023],
        )
        dir_path = op.join(
            self.temp_dir,
            "buoy_data",
            "41001",
            "directional_spectra",
        )
        self.assertTrue(op.exists(dir_path))
        # Check for at least one coefficient file
        coeff_files = list(Path(dir_path).glob("41001*2023.txt.gz"))
        self.assertTrue(len(coeff_files) > 0)
        print(f"\nDirectional spectra files exist: {coeff_files}")


if __name__ == "__main__":
    unittest.main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path as op
import numpy as np
import matplotlib.pyplot as plt

from ..util.time_operations import date2datenum as d2d

# fig aspect and size
_faspect = (1+5**0.5)/2.0
_fsize = 7


# TODO: DEVELOP

def Plot_Hydrographs(dict_bins, p_export=None):
    ''

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    #plt.title('t')
    #plt.xlabel('x')
    #plt.ylabel('y')

    # show / export
    if not p_export:
        plt.show()
    else:
        fig.savefig(p_export, dpi=128)
        plt.close()

def Plot_Hist_TAU(dict_bins, p_export=None):
    ''

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    #plt.title('t')
    #plt.xlabel('x')
    #plt.ylabel('y')

    # show / export
    if not p_export:
        plt.show()
    else:
        fig.savefig(p_export, dpi=128)
        plt.close()

def Plot_Hist_MU(dict_bins, p_export=None):
    ''

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    #plt.title('t')
    #plt.xlabel('x')
    #plt.ylabel('y')

    # show / export
    if not p_export:
        plt.show()
    else:
        fig.savefig(p_export, dpi=128)
        plt.close()

def Plot_Scatter_MU_TAU(dict_bins, p_export=None):
    ''

    # plot figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    #plt.title('t')
    #plt.xlabel('x')
    #plt.ylabel('y')

    # show / export
    if not p_export:
        plt.show()
    else:
        fig.savefig(p_export, dpi=128)
        plt.close()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import scipy.io as sio
from scipy.io.matlab.mio5_params import mat_struct
from scipy.interpolate import griddata
import h5py
import xarray as xr
import numpy as np

# tk
from ..util.time_operations import DateConverter_Mat2Py

def ReadMatfile(p_mfile):
    'Parse .mat file to nested python dictionaries'

    def RecursiveMatExplorer(mstruct_data):
        # Recursive function to extrat mat_struct nested contents

        if isinstance(mstruct_data, mat_struct):
            # mstruct_data is a matlab structure object, go deeper
            d_rc = {}
            for fn in mstruct_data._fieldnames:
                d_rc[fn] = RecursiveMatExplorer(getattr(mstruct_data, fn))
            return d_rc

        else:
            # mstruct_data is a numpy.ndarray, return value
            return mstruct_data

    # base matlab data will be in a dict
    mdata = sio.loadmat(p_mfile, squeeze_me=True, struct_as_record=False)
    mdata_keys = [x for x in mdata.keys() if x not in
                  ['__header__','__version__','__globals__']]

    # use recursive function
    dout = {}
    for k in mdata_keys:
        dout[k] = RecursiveMatExplorer(mdata[k])
    return dout

def ReadGowMat(p_mfile):
    'Read data from gow.mat file. Return xarray.Dataset'

    d_matf = ReadMatfile(p_mfile)

    # parse matlab datenum to datetime
    time = DateConverter_Mat2Py(d_matf['time'])

    # separate keys
    ks_coords = ['time']
    ks_attrs = ['lat','lon','bat','forcing','mesh']

    # generate dataset
    xds_out = xr.Dataset(
        {
        },
        coords = {
            'time': time
        },
        attrs = {
        }
    )
    # fill dataset
    for k in d_matf.keys():
        if k in ks_attrs:
            xds_out.attrs[k] = d_matf[k]
        elif k not in ['time']:
            xds_out[k] =(('time',), d_matf[k])

    return xds_out

def ReadMareografoMat(p_mfile):
    'Read data from Mareografo_KWA.mat file. Return xarray.Dataset'

    d_matf = ReadMatfile(p_mfile)

    # parse matlab datenum to datetime
    time = DateConverter_Mat2Py(d_matf['DATES'])

    # separate keys
    ks_coords = ['time']
    ks_attrs = []

    # generate dataset
    xds_out = xr.Dataset(
        {
        },
        coords = {
            'time': time
        },
        attrs = {
        }
    )
    # fill dataset
    for k in d_matf.keys():
        if k in ks_attrs:
            xds_out.attrs[k] = d_matf[k]
        elif k not in ['DATES']:
            xds_out[k] =(('time',), d_matf[k])

    return xds_out

def ReadCoastMat(p_mfile):
    '''
    Read coast polygons from Costa.mat file.
    Return list of NX2 np.array [x,y]
    '''

    d_matf = ReadMatfile(p_mfile)
    l_pol = []
    for ms in d_matf['costa']:
        l_pol.append(np.array([ms.x, ms.y]).T)
    return l_pol

def ReadEstelaMat(p_mfile):
    '''
    Read estela data from .mat file.
    Return xarray.Dataset
    '''

    threshold = 0

    with h5py.File(p_mfile, 'r') as mf:
        if 'obj' in mf.keys():
            mf = mf['obj']

        # mesh
        mesh_lon = mf['TP']['fullX_centred'][:]
        mesh_lat = mf['full']['Y'][:]
        coast = mf['coastcntr']

        mesh_lon[mesh_lon<0]=mesh_lon[mesh_lon<0] + 360
        longitude = mesh_lon[0,:]
        latitude = mesh_lat[:,0]

        # estela point
        lon_pnt = mf['TP']['lonp'][:]
        lat_pnt = mf['TP']['latp'][:]

        # simple estela land mask
        est_D = mf['C']['traveldays_interp']['y1993to2012'][:]
        mask_land = np.ones(est_D.shape) * np.nan
        mask_land[np.where(est_D<1000000000)]=1

        # mask_shadow (islands), from polar data
        pol_X = mf['Polar']['Xcntr'][:]
        pol_Y = mf['Polar']['Y'][:]
        pol_Fe = mf['C']['Polar_FEmedia']['y1993to2012'][:]

        # interpolate polar to 2D grid to generate mask_shadow
        interp_in = np.column_stack((pol_X.ravel(),pol_Y.ravel()))
        interp_out = pol_Fe.ravel()
        grid_x, grid_y = np.meshgrid(longitude, latitude)
        mask_shadow = griddata(interp_in, interp_out, (grid_x, grid_y), method='nearest')
        mask_shadow[~np.isnan(mask_shadow)]=1

        # generate mask for 0.95 energy
        pol_Fe_95 = pol_Fe.copy()
        pol_Fe_cs = np.nancumsum(pol_Fe, axis=0)
        val95 = pol_Fe_cs[-1]*0.95
        for c in range(pol_Fe_cs.shape[1]):
            ix_del = np.where(pol_Fe_cs[:,c]>=val95[c])[0][0]
            pol_Fe_95[ix_del:,c] = np.nan

        interp_out = pol_Fe_95.ravel()
        grid_x, grid_y = np.meshgrid(longitude, latitude)
        mask_e95 = griddata(interp_in, interp_out, (grid_x, grid_y), method='nearest')
        mask_e95[~np.isnan(mask_e95)]=1

        # fields
        d_D = {}
        d_F = {}
        d_Fthreas = {}
        fds = mf['C']['traveldays_interp'].keys()
        for fd in fds:
            d_D[fd] = mf['C']['traveldays_interp'][fd][:]
            d_F[fd] = mf['C']['FEmedia_interp'][fd][:]
            d_Fthreas[fd] = d_F[fd] / np.nanmax(d_F[fd])

            # use threshold
            d_D[fd][d_Fthreas[fd]<threshold/100] = np.nan
            d_F[fd][d_Fthreas[fd]<threshold/100] = np.nan
            d_Fthreas[fd][d_Fthreas[fd]<threshold/100] = np.nan

    # return xarray.Dataset
    xdset = xr.Dataset(
        {
            'mask_land':(('latitude','longitude'), mask_land),
            'mask_shadow':(('latitude','longitude'), mask_shadow),
            'mask_e95':(('latitude','longitude'), mask_e95),
        },
        coords = {
            'longitude': longitude,
            'latitude': latitude,
        },
        attrs = {
            'pnt_longitude': lon_pnt,
            'pnt_latitude': lat_pnt,
        }
    )

    for k in d_D.keys():
        xdset.update({
            'D_{0}'.format(k):(('latitude','longitude'), d_D[fd]),
            'F_{0}'.format(k):(('latitude','longitude'), d_F[fd]),
            'Fthreas_{0}'.format(k):(('latitude','longitude'), d_Fthreas[fd])
        })

    return xdset

def ReadNakajoMats(p_mfiles):
    '''
    Read Nakajo simulated hurricanes data from .mat files folder.
    Return xarray.Dataset
    '''

    n_sim = 10
    max_ts_end = 0
    n_storms = 0
    var_names = [
        'yts', 'ylon_TC' , 'ylat_TC', 'yDIR', 'ySPEED','yCPRES','del_reason'
    ]

    # generate var holder dict
    d_vars = {}
    for vn in var_names:
        d_vars[vn] = np.array([])

    # find number and time length of synthetic storms
    for i in range(n_sim):

        # read sim file
        p_matf = op.join(p_mfiles, 'YCAL{0}.mat'.format(i+1))
        d_matf = ReadMatfile(p_matf)

        # count storms
        n_storms += len(d_matf[var_names[0]])

        # append data to var holder
        for vn in var_names:
            d_vars[vn] = np.concatenate((d_vars[vn], d_matf[vn]))

    # add to xarray dataset
    xds_out = xr.Dataset(
        {
            'yts':(('storm',), d_vars['yts']),
            'ylon_TC':(('storm',), d_vars['ylon_TC']),
            'ylat_TC':(('storm',), d_vars['ylat_TC']),
            'yDIR':(('storm',), d_vars['yDIR']),
            'ySPEED':(('storm',), d_vars['ySPEED']),
            'yCPRES':(('storm',), d_vars['yCPRES']),
            'del_reason':(('storm',), d_vars['del_reason']),
        },
        coords = {
            'storm':range(n_storms)
        }
    )
    return xds_out

def ReadTCsSimulations(p_sims, point_file='Punto1.mat'):
    '''
    Read data from solved TCs simulations. Return xarray.Dataset

    TCs simulations are stored at folders named "CYC_numbercode"
    '''

    # get all folders at TCs simulations folder
    f_sims = sorted([x for x in os.listdir(p_sims) if op.isdir(op.join(p_sims, x))])

    # ensure number order
    ens_order = np.argsort([int(x.split('_')[-1]) for x in f_sims])
    f_sims_e = [f_sims[x] for x in ens_order]

    # now make full path to CYC folders
    sims = [op.join(p_sims, x) for x in f_sims_e]

    # storm parameters
    n_sims = len(sims)
    xds_storms_sims = xr.Dataset(
        {
            'hs':(('storm',), np.ones(n_sims) * np.nan),
            'tp':(('storm',), np.ones(n_sims) * np.nan),
            'dir':(('storm',), np.ones(n_sims) * np.nan),
            'ss':(('storm',), np.ones(n_sims) * np.nan),
            'twl':(('storm',), np.ones(n_sims) * np.nan),
            'mu':(('storm',), np.ones(n_sims) * np.nan),
        },
        coords = {'storm':np.arange(n_sims)}
    )

    # read file by file
    for i_s, p_s in enumerate(sims):
        p_mat = op.join(p_s, point_file)

        d_mat = ReadMatfile(p_mat)
        data = d_mat['data']
        hs = data['hs']
        tp = data['tp']
        dr = data['dir']
        ss = data['ss'] / 100.0
        time = data['time']

        # calculate TWL
        twl = 0.053 * hs**0.5 * tp + ss

        # calculate max TWL position
        i_max = np.nanargmax(twl)
        twl_max = np.nanmax(twl)

        # calculate MU (12 - 12 hours window)
        pos = np.arange(len(ss))
        ix = np.where((pos>=i_max-12) & (pos<=i_max+12))
        twl_w = twl[ix]

        dt = 1.0 / len(twl_w)
        time_norm = np.arange(dt,1+dt,dt)
        mu = np.trapz(twl_w, time_norm) / twl_max

        # store storm parameters
        xds_storms_sims.hs[i_s] = hs[i_max]
        xds_storms_sims.tp[i_s] = tp[i_max]
        xds_storms_sims.dir[i_s] = dr[i_max]
        xds_storms_sims.ss[i_s] = ss[i_max]
        xds_storms_sims.twl[i_s] = twl[i_max]
        xds_storms_sims.mu[i_s] = mu

    return xds_storms_sims




"""
Module attrs
"""

__version__     = '0.1.0'
__author__      = 'GeoOcean(UC)'
__contact__     = 'ripolln@unican.es'
__url__         = 'https://gitlab.com/geoocean/bluemath/mw_deltares/'
__description__ = 'SWAN numerical model python wrap'
__keywords__    = 'SWAN waves numerical'

import os.path as op
import shutil

from . import geo
from . import io
from . import storms
from . import plots
from . import wrap



def set_swan_binary_file(bin_file):
    'copy swan bin_file to wrap bin location'

    # swan bin executable
    p_res = op.join(op.dirname(op.realpath(__file__)), 'resources')
    p_bin = op.abspath(op.join(p_res, 'swan_bin', 'swan_ser.exe'))

    # copy file
    shutil.copyfile(bin_file, p_bin)

    print('bin file copied to {0}'.format(p_bin))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common 
import os
import os.path as op

# pip
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import path

# tk 
from .pca import PCA_EstelaPred
from .kma import KMA_regression_guided
from .kma import SimpleMultivariateRegressionModel as SMRM
from .intradaily import Calculate_Hydrographs

from .plotting.estela import Plot_EOFs_EstelaPred, Plot_DWTs_Mean_Anom, \
Plot_DWTs_Probs
from .plotting.pcs import Plot_PCs_WT, Plot_WT_PCs_3D


def spatial_gradient(xdset, var_name):
    '''
    Calculate spatial gradient

    xdset:
        (longitude, latitude, time), var_name

    returns xdset with new variable "var_name_gradient"
    '''

    # TODO:check/ ADD ONE ROW/COL EACH SIDE
    var_grad = np.zeros(xdset[var_name].shape)

    Mx = len(xdset.longitude)
    My = len(xdset.latitude)
    lat = xdset.latitude.values

    for it in range(len(xdset.time)):
        var_val = xdset[var_name].isel(time=it).values

        # calculate gradient (matrix)
        m_c = var_val[1:-1,1:-1]
        m_l = np.roll(var_val, -1, axis=1)[1:-1,1:-1]
        m_r = np.roll(var_val, +1, axis=1)[1:-1,1:-1]
        m_u = np.roll(var_val, -1, axis=0)[1:-1,1:-1]
        m_d = np.roll(var_val, +1, axis=0)[1:-1,1:-1]
        m_phi = np.pi*np.abs(lat)/180.0
        m_phi = m_phi[1:-1]

        dpx1 = (m_c - m_l)/np.cos(m_phi[:,None])
        dpx2 = (m_r - m_c)/np.cos(m_phi[:,None])
        dpy1 = m_c - m_d
        dpy2 = m_u - m_c

        vg = (dpx1**2+dpx2**2)/2 + (dpy1**2+dpy2**2)/2
        var_grad[it, 1:-1, 1:-1] = vg

        # calculate gradient (for). old code
        #for i in range(1, Mx-1):
        #    for j in range(1, My-1):
        #        phi = np.pi*np.abs(lat[j])/180.0
        #        dpx1 = (var_val[j,i]   - var_val[j,i-1]) / np.cos(phi)
        #        dpx2 = (var_val[j,i+1] - var_val[j,i])   / np.cos(phi)
        #        dpy1 = (var_val[j,i]   - var_val[j-1,i])
        #        dpy2 = (var_val[j+1,i] - var_val[j,i])
        #        var_grad[it, j, i] = (dpx1**2+dpx2**2)/2 + (dpy1**2+dpy2**2)/2

    # store gradient
    xdset['{0}_gradient'.format(var_name)]= (
        ('time', 'latitude', 'longitude'), var_grad)

    return xdset

def mask_from_poly(xdset, ls_poly, name_mask='mask'):
    '''
    Generate mask from list of tuples (lon, lat)

    xdset dimensions:
        (longitude, latitude, )

    returns xdset with new variable "mask"
    '''

    lon = xdset.longitude.values
    lat = xdset.latitude.values
    mesh_lat, mesh_lon = np.meshgrid(lat, lon)
    mask = np.zeros(mesh_lat.shape)

    mesh_points = np.array(
        [mesh_lon.flatten(), mesh_lat.flatten()]
    ).T

    for pol in ls_poly:
        p = path.Path(pol)
        inside = np.array(p.contains_points(mesh_points))
        inmesh = np.reshape(inside, mask.shape)
        mask[inmesh] = 1

    xdset[name_mask]=(('latitude','longitude'), mask.T)

    return xdset

def dynamic_estela_predictor(xdset, var_name, estela_D):
    '''
    Generate dynamic predictor using estela

    xdset:
        (time, latitude, longitude), var_name, mask

    returns similar xarray.Dataset with variables:
        (time, latitude, longitude), var_name_comp
        (time, latitude, longitude), var_name_gradient_comp
    '''

    # first day is estela max
    first_day = int(np.floor(np.nanmax(estela_D)))+1

    # output will start at time=first_day
    shp = xdset[var_name].shape
    comp_shape = (shp[0]-first_day, shp[1], shp[2])
    var_comp = np.ones(comp_shape) * np.nan
    var_grd_comp = np.ones(comp_shape) * np.nan

    # get data using estela for each cell
    for i_lat in range(len(xdset.latitude)):
        for i_lon in range(len(xdset.longitude)):
            ed = estela_D[i_lat, i_lon]
            if not np.isnan(ed):

                # mount estela displaced time array 
                i_times = np.arange(
                    first_day, len(xdset.time)
                ) - np.int(ed)

                # select data from displaced time array positions
                xdselec = xdset.isel(
                    time = i_times,
                    latitude = i_lat,
                    longitude = i_lon)

                # get estela predictor values
                var_comp[:, i_lat, i_lon] = xdselec[var_name].values
                var_grd_comp[:, i_lat, i_lon] = xdselec['{0}_gradient'.format(var_name)].values

    # return generated estela predictor
    return xr.Dataset(
        {
            '{0}_comp'.format(var_name):(
                ('time','latitude','longitude'), var_comp),
            '{0}_gradient_comp'.format(var_name):(
                ('time','latitude','longitude'), var_grd_comp),

        },
        coords = {
            'time':xdset.time.values[first_day:],
            'latitude':xdset.latitude.values,
            'longitude':xdset.longitude.values,
        }
    )


class Predictor(object):
    '''
    tesla-kit custom dataset handler

    used for 3D dataset (lon,lat,time) and related
    statistical classification calculations and figures.
    '''

    def __init__(self, p_store):

        # file paths
        self.p_store = p_store
        self.p_data = op.join(p_store, 'data.nc')
        self.p_pca = op.join(p_store, 'pca.nc')
        self.p_kma = op.join(p_store, 'kma.nc')
        self.p_plots = op.join(p_store, 'figs')

        # data (xarray.Dataset)
        self.data = None
        self.PCA = None
        self.KMA = None

    def Load(self):
        if op.isfile(self.p_data):
            self.data = xr.open_dataset(self.p_data)
        if op.isfile(self.p_pca):
            self.PCA = xr.open_dataset(self.p_pca)
        if op.isfile(self.p_kma):
            self.KMA = xr.open_dataset(self.p_kma)

    def Save(self):
        try:
            os.makedirs(self.p_store)
        except:
            pass
        if self.data:
            if op.isfile(self.p_data): os.remove(self.p_data)
            self.data.to_netcdf(self.p_data,'w')
        if self.PCA:
            if op.isfile(self.p_pca): os.remove(self.p_pca)
            self.PCA.to_netcdf(self.p_pca,'w')
        if self.KMA:
            if op.isfile(self.p_kma): os.remove(self.p_kma)
            self.KMA.to_netcdf(self.p_kma,'w')

    def Calc_PCA_EstelaPred(self, var_name, xds_estela):
        'Principal components analysis using estela predictor'

        # generate estela predictor
        # TODO STORE DYNAMIC ESTELA PREDICTOR (too big..)
        xds_estela_pred = dynamic_estela_predictor(
            self.data, var_name, xds_estela)

        # Calculate PCA
        self.PCA = PCA_EstelaPred(
            xds_estela_pred, var_name)

        # save data
        self.Save()

    def Calc_KMA_regressionguided(
        self, num_clusters, xds_waves, waves_vars, alpha, min_group_size=None, repres = 0.95):
        'KMA regression guided with waves data'

        # we have to miss some days of data due to ESTELA
        tcut = self.PCA.pred_time.values[:]

        # calculate regresion model between predictand and predictor
        xds_waves = xds_waves.sel(time = slice(tcut[0], tcut[-1]))
        xds_Yregres = SMRM(self.PCA, xds_waves, waves_vars)

        # classification: KMA regresion guided
        self.KMA = KMA_regression_guided(
            self.PCA, xds_Yregres,
            num_clusters, repres, alpha, min_group_size
        )

        # store time array with KMA
        self.KMA['time'] = (('n_components',), self.PCA.pred_time.values[:])

        # save data
        self.Save()

    def Calc_MU_TAU_Hydrographs(self, xds_WAVES):
        '''
        Calculates TWL hydrographs

        returns list of xarray.Dataset with TWL hydrographs MU,TAU arrays for each WT
        '''

        # TODO: SACAR DE AQUI, replantear hydrographs

        # get sorted bmus from kma
        xds_BMUS = xr.Dataset(
            {'bmus':(('time', self.KMA.sorted_bmus.values[:]))},
            coords = {'time': self.KMA.time.values[:]}
        )

        # Calculate hydrographs for each WT
        _, l_xds_MUTAU = Calculate_Hydrographs(xds_BMUS, xds_WAVES)

        return l_xds_MUTAU

    def Mod_KMA_AddStorms(self, storm_dates, storm_categories):
        '''
        Modify KMA bmus series adding storm category (6 new groups)
        '''

        n_clusters = len(self.KMA.n_clusters.values[:])
        kma_dates = self.KMA.time.values[:]
        bmus_storms = np.copy(self.KMA.sorted_bmus.values[:])  # copy numpy.array

        for sd, sc in zip(storm_dates, storm_categories):
            sdr =  np.array(sd, dtype='datetime64[D]')  # round to day
            pos_date = np.where(kma_dates==sdr)[0]
            if pos_date:
                bmus_storms[pos_date[0]] = n_clusters + sc

        # copy kma and add bmus_storms
        self.KMA['sorted_bmus_storms'] = (('n_components',), bmus_storms)

        # store changes
        if op.isfile(self.p_kma): os.remove(self.p_kma)
        self.KMA.to_netcdf(self.p_kma,'w')

    def Plot_EOFs_EstelaPred(self, n_plot=3, show=True):
        'Plot EOFs generated in PCA_EstelaPred'

        # optional land mask
        mask_land = None
        if 'mask_land' in self.data.keys():
            mask_land = self.data['mask_land'].values[:]

        l_fs = Plot_EOFs_EstelaPred(self.PCA, n_plot, mask_land=mask_land, show=show)
        return l_fs

    def Plot_DWTs(self, var_name, kind='mean', show=True):
        '''
        Plot KMA clusters generated in PCA_EstelaPred (DWTs)

        kind = mean / anom

        uses database means/anomalies at cluster location (bmus corrected)
        '''

        # data to plot
        xds_DWTs = self.KMA
        var_data = self.data[var_name]

        # optional land mask
        mask_land = None
        if 'mask_land' in self.data.keys():
            mask_land = self.data['mask_land'].values[:]

        # Plot DWTs mean using var_data
        fig = Plot_DWTs_Mean_Anom(
            xds_DWTs, var_data,
            kind = kind,
            mask_land = mask_land,
            show = show
        )

        return fig

    def Plot_DWTs_Probs(self, show=True, field='sorted_bmus', n_clusters=None):
        '''
        Plot DWTs bmus probabilities
            - histogram for ocurrences
            - probs. all series
            - probs by month
            - probs by 3month
        '''

        # Plot DWTs mean using var_data
        bmus = self.KMA[field].values[:] + 1 # index to DWT id
        bmus_time = self.KMA['time'].values[:]
        if not n_clusters:
            n_clusters = len(self.KMA.n_clusters.values[:])

        fig = Plot_DWTs_Probs(bmus, bmus_time, n_clusters, show=show)

        return fig

    def Plot_DWT_PCs(self, n=3, show=True):
        '''
        Plot Daily Weather Types PCs using 2D axis
        '''

        # get estela data
        PCs = self.PCA.PCs.values[:]
        variance = self.PCA.variance.values[:]
        bmus = self.KMA.sorted_bmus.values[:]  # sorted_bmus
        n_clusters = len(self.KMA.n_clusters.values[:])

        # Plot DWTs PCs
        fig = Plot_PCs_WT(PCs, variance, bmus, n_clusters, n, show=show)

        return fig

    def Plot_PCs_3D(self, show=True):
        'Plots Predictor first 3 PCs'

        # first 3 PCs
        bmus = self.KMA['sorted_bmus'].values[:]
        PCs = self.PCA.PCs.values[:]
        variance = self.PCA.variance.values[:]

        n_clusters = len(self.KMA.n_clusters.values[:])

        PC1 = np.divide(PCs[:,0], np.sqrt(variance[0]))
        PC2 = np.divide(PCs[:,1], np.sqrt(variance[1]))
        PC3 = np.divide(PCs[:,2], np.sqrt(variance[2]))

        # dictionary of DWT PCs 123 
        d_PCs = {}
        for ic in range(n_clusters):
            ind = np.where(bmus == ic)[:]

            PC123 = np.column_stack((PC1[ind], PC2[ind], PC3[ind]))
            d_PCs['{0}'.format(ic+1)] = PC123

        # Plot DWTs PCs 3D 
        fig = Plot_WT_PCs_3D(d_PCs, n_clusters, show=show)

        return fig




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import collections
import numpy as np
from math import sqrt


def GetDivisors(x):
    l_div = []
    i = 1
    while i<x:
        if x%i == 0:
            l_div.append(i)
        i = i + 1
    return l_div

def GetUniqueRows(np_array):
    d = collections.OrderedDict()
    for a in np_array:
        t = tuple(a)
        if t in d:
            d[t] += 1
        else:
            d[t] = 1

    result = []
    for (key, value) in d.items():
        result.append(list(key) + [value])

    np_result = np.asarray(result)
    return np_result

def GetBestRowsCols(n):
    'try to square number n, used at gridspec plots'

    sqrt_n = sqrt(n)
    if sqrt_n.is_integer():
        n_r = int(sqrt_n)
        n_c = int(sqrt_n)
    else:
        l_div = GetDivisors(n)
        n_c = l_div[int(len(l_div)/2)]
        n_r = int(n/n_c)

    return n_r, n_c

def GetRepeatedValues(series):
    'Find adyacent repeated values inside series. Return list of tuples'

    ix = 0
    s0, s1 = None, None
    l_subseq_index = []
    while ix < len(series)-1:

        # subsequence start
        if series[ix] == series[ix+1] and s0==None: s0 = ix

        # subsequence end
        elif series[ix] != series[ix+1] and s0!=None: s1 = ix + 1

        # series end
        if ix == len(series)-2: s1 = ix + 2

        # store subsequence
        if s0!=None and s1!=None:
            l_subseq_index.append((s0, s1))
            s0, s1 = None, None

        ix+=1

    return l_subseq_index




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
import numpy as np

def Normalize(data, ix_scalar, ix_directional, minis=[], maxis=[]):
    '''
    Normalize data subset - norm = val - min) / (max - min)

    data - data to normalize, data variables at columns.
    ix_scalar - scalar columns indexes
    ix_directional - directional columns indexes
    '''

    data_norm = np.zeros(data.shape) * np.nan

    # calculate maxs and mins 
    if minis==[] or maxis==[]:

        # scalar data
        for ix in ix_scalar:
            v = data[:, ix]
            mi = np.amin(v)
            ma = np.amax(v)
            data_norm[:, ix] = (v - mi) / (ma - mi)
            minis.append(mi)
            maxis.append(ma)

        minis = np.array(minis)
        maxis = np.array(maxis)

    # max and mins given
    else:

        # scalar data
        for c, ix in enumerate(ix_scalar):
            v = data[:, ix]
            mi = minis[c]
            ma = maxis[c]
            data_norm[:,ix] = (v - mi) / (ma - mi)

    # directional data
    for ix in ix_directional:
        v = data[:,ix]
        data_norm[:,ix] = v * np.pi / 180.0


    return data_norm, minis, maxis

def DeNormalize(data_norm, ix_scalar, ix_directional, minis, maxis):
    '''
    DeNormalize data subset for MaxDiss algorithm

    data - data to normalize, data variables at columns.
    ix_scalar - scalar columns indexes
    ix_directional - directional columns indexes
    '''

    data = np.zeros(data_norm.shape) * np.nan

    # scalar data
    for c, ix in enumerate(ix_scalar):
        v = data_norm[:,ix]
        mi = minis[c]
        ma = maxis[c]
        data[:, ix] = v * (ma - mi) + mi

    # directional data
    for ix in ix_directional:
        v = data_norm[:,ix]
        data[:, ix] = v * 180 / np.pi

    return data

def Normalized_Distance(M, D, ix_scalar, ix_directional):
    '''
    Normalized distance between rows in M and D

    M - numpy array
    D - numpy array
    ix_scalar - scalar columns indexes
    ix_directional - directional columns indexes
    '''

    dif = np.zeros(M.shape)

    # scalar
    for ix in ix_scalar:
        dif[:,ix] = D[:,ix] - M[:,ix]

    # directional
    for ix in ix_directional:
        ab = np.absolute(D[:,ix] - M[:,ix])
        dif[:,ix] = np.minimum(ab, 2*np.pi - ab)/np.pi

    dist = np.sum(dif**2,1)
    return dist

def nearest_indexes(data_q, data, ix_scalar, ix_directional):
    '''
    for each row in data_q, find nearest point in data and store index.

    Returns array of indexes of each nearest point to all entries in data_q
    '''

    # normalize scalar and directional data 
    data_norm, minis, maxis = Normalize(data, ix_scalar, ix_directional)
    data_q_norm, _, _ = Normalize(
        data_q, ix_scalar, ix_directional,
        minis=minis, maxis=maxis
    )

    # compute distances, store nearest distance index
    ix_near = np.zeros(data_q_norm.shape[0]).astype(int)
    for c, dq in enumerate(data_q_norm):
        ddq = np.repeat([dq], data_norm.shape[0], axis=0)
        D = Normalized_Distance(data_norm, ddq, ix_scalar, ix_directional)
        ix_near[c] = np.argmin(D)

    return ix_near

def MaxDiss_Simplified_NoThreshold(data, num_centers, ix_scalar, ix_directional):
    '''
    Normalize data and calculate centers using
    maxdiss simplified no-threshold algorithm

    data - data to apply maxdiss algorithm, data variables at columns
    num_centers - number of centers to calculate
    ix_scalar - scalar columns indexes
    ix_directional - directional columns indexes
    '''

    # TODO: REFACTOR / OPTIMIZE 

    print('\nMaxDiss dataset: {0} --> {1}\n'.format(
        data.shape[0], num_centers))

    # normalize scalar and directional data
    data_norm, minis, maxis = Normalize(data, ix_scalar, ix_directional)

    # mda seed
    seed = np.where(data_norm[:,0] == np.amax(data_norm[:,0]))[0][0]

    # initialize centroids subset
    subset = np.array([data_norm[seed]])
    train = np.delete(data_norm, seed, axis=0)

    # repeat till we have desired num_centers
    n_c = 1
    while n_c < num_centers:
        m = np.ones((train.shape[0],1))
        m2 = subset.shape[0]

        if m2 == 1:
            xx2 = np.repeat(subset, train.shape[0], axis=0)
            d_last = Normalized_Distance(train, xx2, ix_scalar, ix_directional)

        else:
            xx = np.array([subset[-1,:]])
            xx2 = np.repeat(xx, train.shape[0], axis=0)
            d_prev = Normalized_Distance(train, xx2, ix_scalar, ix_directional)
            d_last = np.minimum(d_prev, d_last)

        qerr, bmu = np.amax(d_last), np.argmax(d_last)

        if not np.isnan(qerr):
            subset = np.append(subset, np.array([train[bmu,:]]), axis=0)
            train = np.delete(train, bmu, axis=0)
            d_last = np.delete(d_last, bmu, axis=0)

            # log
            fmt = '0{0}d'.format(len(str(num_centers)))
            print('   MDA centroids: {1:{0}}/{2:{0}}'.format(
                fmt, subset.shape[0], num_centers), end='\r')

        n_c = subset.shape[0]
    print('\n')

    # normalize scalar and directional data
    centroids = DeNormalize(subset, ix_scalar, ix_directional, minis, maxis)

    return centroids




from typing import List, Tuple

import numpy as np


def upcrossing(series: np.ndarray) -> Tuple[List[float], np.ndarray]:
    """
    Calculate wave heights and periods using the zero-crossing method.

    This function performs wave analysis by detecting zero-crossings in a surface
    elevation time series. It removes the mean from the series and identifies
    upcrossings (where the signal crosses zero from negative to positive).
    For each wave cycle between upcrossings, it calculates the wave height
    (difference between maximum and minimum) and wave period (time between crossings).

    Parameters
    ----------
    series : np.ndarray
        A 2D array with shape (n_samples, 2) where:
        - series[:, 0] contains the time values
        - series[:, 1] contains the surface elevation values

    Returns
    -------
    Tuple[List[float], np.ndarray]
        A tuple containing:
        - Hi : List of wave heights (amplitudes) for each wave cycle
        - Ti : Array of wave periods (time intervals between zero crossings)

    Notes
    -----
    - The function removes the mean from the surface elevation series before analysis
    - Wave heights are calculated as the difference between maximum and minimum
      values within each wave cycle
    - Wave periods are the time intervals between consecutive zero upcrossings
    - This method is commonly used in oceanography for wave analysis

    Examples
    --------
    >>> import numpy as np
    >>> time = np.linspace(0, 10, 1000)
    >>> elevation = np.sin(2 * np.pi * 0.5 * time) + 0.1 * np.random.randn(1000)
    >>> series = np.column_stack([time, elevation])
    >>> heights, periods = upcrossing(series)
    """

    # Remove mean from surface elevation series
    series[:, 1] -= np.mean(series[:, 1])

    # Find zero crossings (where signal changes from negative to positive)
    zeros = np.where(np.diff(np.sign(series[:, 1])) > 0)[0]

    # Calculate wave periods (time intervals between zero crossings)
    Ti = np.diff(zeros)

    # Calculate wave heights (amplitudes) for each wave cycle
    Hi = [
        np.max(series[:, 1][zeros[i] : zeros[i + 1]])
        - np.min(series[:, 1][zeros[i] : zeros[i + 1]])
        for i in range(len(zeros) - 1)
    ]

    return Hi, Ti



import unittest

import numpy as np
import pandas as pd
import xarray as xr
from sklearn.preprocessing import StandardScaler

from bluemath_tk.core.operations import (
    convert_lonlat_to_utm,
    convert_utm_to_lonlat,
    denormalize,
    destandarize,
    get_degrees_from_uv,
    get_uv_components,
    mathematical_to_nautical,
    nautical_to_mathematical,
    normalize,
    spatial_gradient,
    standarize,
)


class TestNormalize(unittest.TestCase):
    """Test the normalize function."""

    def setUp(self):
        """Set up test data."""

        self.df_data = pd.DataFrame(
            {
                "Hs": [1.0, 2.0, 3.0, 4.0, 5.0],
                "Tp": [5.0, 10.0, 15.0, 20.0, 25.0],
                "Dir": [0.0, 90.0, 180.0, 270.0, 360.0],
            }
        )
        self.ds_data = xr.Dataset(
            {
                "Hs": (("time",), [1.0, 2.0, 3.0, 4.0, 5.0]),
                "Tp": (("time",), [5.0, 10.0, 15.0, 20.0, 25.0]),
                "Dir": (("time",), [0.0, 90.0, 180.0, 270.0, 360.0]),
            },
            coords={"time": pd.date_range("2000-01-01", periods=5)},
        )

    def test_normalize_dataframe(self):
        """Test normalization of pandas DataFrame."""

        normalized_data, scale_factor = normalize(self.df_data)

        # Check that data is normalized to 0-1 range
        for col in normalized_data.columns:
            self.assertAlmostEqual(normalized_data[col].min(), 0.0, places=10)
            self.assertAlmostEqual(normalized_data[col].max(), 1.0, places=10)

        # Check scale factor structure
        for var in scale_factor:
            self.assertIn(var, self.df_data.columns)
            self.assertEqual(len(scale_factor[var]), 2)
            self.assertLess(scale_factor[var][0], scale_factor[var][1])

    def test_normalize_dataset(self):
        """Test normalization of xarray Dataset."""

        normalized_data, scale_factor = normalize(self.ds_data)

        # Check that data is normalized to 0-1 range
        for var in normalized_data.data_vars:
            self.assertAlmostEqual(normalized_data[var].min(), 0.0, places=10)
            self.assertAlmostEqual(normalized_data[var].max(), 1.0, places=10)

        # Check scale factor structure
        for var in scale_factor:
            self.assertIn(var, self.ds_data.data_vars)
            self.assertEqual(len(scale_factor[var]), 2)

    def test_normalize_with_custom_scale_factor(self):
        """Test normalization with custom scale factors."""

        custom_scale = {"Hs": [0.0, 10.0], "Tp": [0.0, 30.0], "Dir": [0.0, 360.0]}
        _normalized_data, scale_factor = normalize(self.df_data, custom_scale)

        # Check that custom scale factors are used
        self.assertEqual(scale_factor["Hs"], [0.0, 10.0])
        self.assertEqual(scale_factor["Tp"], [0.0, 30.0])
        self.assertEqual(scale_factor["Dir"], [0.0, 360.0])

    def test_normalize_invalid_data_type(self):
        """Test that invalid data type raises TypeError."""

        with self.assertRaises(TypeError):
            normalize(np.array([1, 2, 3]))

    def test_normalize_preserves_original_data(self):
        """Test that original data is not modified."""

        original_data = self.df_data.copy()
        normalize(self.df_data)
        pd.testing.assert_frame_equal(original_data, self.df_data)


class TestDenormalize(unittest.TestCase):
    """Test the denormalize function."""

    def setUp(self):
        """Set up test data."""

        self.normalized_df = pd.DataFrame(
            {"Hs": [0.0, 0.25, 0.5, 0.75, 1.0], "Tp": [0.0, 0.25, 0.5, 0.75, 1.0]}
        )
        self.normalized_ds = xr.Dataset(
            {
                "Hs": (("time",), [0.0, 0.25, 0.5, 0.75, 1.0]),
                "Tp": (("time",), [0.0, 0.25, 0.5, 0.75, 1.0]),
            },
            coords={"time": pd.date_range("2000-01-01", periods=5)},
        )
        self.scale_factor = {"Hs": [1.0, 5.0], "Tp": [5.0, 25.0]}

    def test_denormalize_dataframe(self):
        """Test denormalization of pandas DataFrame."""

        denormalized_data = denormalize(self.normalized_df, self.scale_factor)

        # Check that data is denormalized correctly
        expected_hs = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        expected_tp = np.array([5.0, 10.0, 15.0, 20.0, 25.0])

        np.testing.assert_allclose(denormalized_data["Hs"].values, expected_hs)
        np.testing.assert_allclose(denormalized_data["Tp"].values, expected_tp)

    def test_denormalize_dataset(self):
        """Test denormalization of xarray Dataset."""

        denormalized_data = denormalize(self.normalized_ds, self.scale_factor)

        # Check that data is denormalized correctly
        expected_hs = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        expected_tp = np.array([5.0, 10.0, 15.0, 20.0, 25.0])

        np.testing.assert_allclose(denormalized_data["Hs"].values, expected_hs)
        np.testing.assert_allclose(denormalized_data["Tp"].values, expected_tp)

    def test_denormalize_invalid_data_type(self):
        """Test that invalid data type raises TypeError."""

        with self.assertRaises(TypeError):
            denormalize(np.array([1, 2, 3]), self.scale_factor)

    def test_denormalize_preserves_original_data(self):
        """Test that original data is not modified."""

        original_data = self.normalized_df.copy()
        denormalize(self.normalized_df, self.scale_factor)
        pd.testing.assert_frame_equal(original_data, self.normalized_df)


class TestStandarize(unittest.TestCase):
    """Test the standarize function."""

    def setUp(self):
        """Set up test data."""

        self.np_data = np.array(
            [
                [1.0, 2.0, 3.0, 4.0, 5.0],
                [5.0, 10.0, 15.0, 20.0, 25.0],
                [0.0, 90.0, 180.0, 270.0, 360.0],
            ]
        )
        self.df_data = pd.DataFrame(
            {
                "Hs": [1.0, 2.0, 3.0, 4.0, 5.0],
                "Tp": [5.0, 10.0, 15.0, 20.0, 25.0],
                "Dir": [0.0, 90.0, 180.0, 270.0, 360.0],
            }
        )
        self.ds_data = xr.Dataset(
            {
                "Hs": (("time",), [1.0, 2.0, 3.0, 4.0, 5.0]),
                "Tp": (("time",), [5.0, 10.0, 15.0, 20.0, 25.0]),
                "Dir": (("time",), [0.0, 90.0, 180.0, 270.0, 360.0]),
            },
            coords={"time": pd.date_range("2000-01-01", periods=5)},
        )

    def test_standarize_numpy_array(self):
        """Test standardization of numpy array."""

        standarized_data, scaler = standarize(self.np_data)

        # Check that data has mean close to 0 and std close to 1
        self.assertAlmostEqual(standarized_data.mean(), 0.0, delta=0.15)
        self.assertAlmostEqual(standarized_data.std(), 1.0, delta=0.15)

        # Check that scaler is StandardScaler instance
        self.assertIsInstance(scaler, StandardScaler)

    def test_standarize_dataframe(self):
        """Test standardization of pandas DataFrame."""

        standarized_data, _scaler = standarize(self.df_data)

        # Check that data has mean close to 0 and std close to 1
        for col in standarized_data.columns:
            self.assertAlmostEqual(standarized_data[col].mean(), 0.0, delta=0.15)
            self.assertAlmostEqual(standarized_data[col].std(), 1.0, delta=0.15)

        # Check that column names are preserved
        self.assertEqual(list(standarized_data.columns), list(self.df_data.columns))

    def test_standarize_dataset(self):
        """Test standardization of xarray Dataset."""

        standarized_data, _scaler = standarize(self.ds_data)

        # Check that data has mean close to 0 and std close to 1
        for var in standarized_data.data_vars:
            self.assertAlmostEqual(standarized_data[var].mean(), 0.0, delta=0.15)
            self.assertAlmostEqual(standarized_data[var].std(), 1.0, delta=0.15)

        # Check that coordinates are preserved
        self.assertEqual(list(standarized_data.coords), list(self.ds_data.coords))

    def test_standarize_with_existing_scaler(self):
        """Test standardization with existing scaler."""

        scaler = StandardScaler()
        _standarized_data, returned_scaler = standarize(self.np_data, scaler=scaler)

        self.assertIs(returned_scaler, scaler)

    def test_standarize_transform_only(self):
        """Test standardization with transform=True."""

        # First fit a scaler
        _, scaler = standarize(self.np_data)

        # Then transform new data
        new_data = np.array(
            [[1, 2, 3, 4, 5], [5, 10, 15, 20, 25], [0, 90, 180, 270, 360]], dtype=float
        )
        standarized_data, _ = standarize(new_data, scaler=scaler, transform=True)

        # Check that data is transformed but not fitted
        self.assertNotEqual(standarized_data.mean(), 0.0)


class TestDestandarize(unittest.TestCase):
    """Test the destandarize function."""

    def setUp(self):
        """Set up test data."""

        self.original_data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=float)
        self.standarized_data, self.scaler = standarize(self.original_data)

    def test_destandarize_numpy_array(self):
        """Test destandarization of numpy array."""

        destandarized_data = destandarize(self.standarized_data, self.scaler)

        # Check that data is restored to original values
        np.testing.assert_allclose(destandarized_data, self.original_data, rtol=1e-10)

    def test_destandarize_dataframe(self):
        """Test destandarization of pandas DataFrame."""

        df_standarized = pd.DataFrame(self.standarized_data, columns=["A", "B", "C"])
        df_original = pd.DataFrame(self.original_data, columns=["A", "B", "C"])

        destandarized_data = destandarize(df_standarized, self.scaler)

        # Check that data is restored to original values
        pd.testing.assert_frame_equal(destandarized_data, df_original)

    def test_destandarize_dataset(self):
        """Test destandarization of xarray Dataset."""

        ds_standarized = xr.Dataset(
            {
                "A": (("time",), self.standarized_data[:, 0]),
                "B": (("time",), self.standarized_data[:, 1]),
                "C": (("time",), self.standarized_data[:, 2]),
            },
            coords={"time": pd.date_range("2000-01-01", periods=3)},
        )
        ds_original = xr.Dataset(
            {
                "A": (("time",), self.original_data[:, 0]),
                "B": (("time",), self.original_data[:, 1]),
                "C": (("time",), self.original_data[:, 2]),
            },
            coords={"time": pd.date_range("2000-01-01", periods=3)},
        )
        destandarized_data = destandarize(ds_standarized, self.scaler)

        # Check that data is restored to original values
        xr.testing.assert_allclose(destandarized_data, ds_original)


class TestUVComponents(unittest.TestCase):
    """Test the get_uv_components and get_degrees_from_uv functions."""

    def test_get_uv_components_scalar(self):
        """Test UV components calculation with scalar input."""

        xu, xv = get_uv_components(np.array([0]))

        # North direction (0°) should give u=0, v=1
        self.assertAlmostEqual(xu[0], 0.0, places=10)
        self.assertAlmostEqual(xv[0], 1.0, places=10)

    def test_get_uv_components_array(self):
        """Test UV components calculation with array input."""

        angles = np.array([0, 90, 180, 270])
        xu, xv = get_uv_components(angles)

        # Check specific directions
        # North (0°)
        self.assertAlmostEqual(xu[0], 0.0, places=10)
        self.assertAlmostEqual(xv[0], 1.0, places=10)

        # East (90°)
        self.assertAlmostEqual(xu[1], 1.0, places=10)
        self.assertAlmostEqual(xv[1], 0.0, places=10)

        # South (180°)
        self.assertAlmostEqual(xu[2], 0.0, places=10)
        self.assertAlmostEqual(xv[2], -1.0, places=10)

        # West (270°)
        self.assertAlmostEqual(xu[3], -1.0, places=10)
        self.assertAlmostEqual(xv[3], 0.0, places=10)

    def test_get_degrees_from_uv_scalar(self):
        """Test degrees calculation from UV components with scalar input."""

        xu = np.array([0])
        xv = np.array([1])
        degrees = get_degrees_from_uv(xu, xv)

        # u=0, v=1 should give 0° (North)
        self.assertAlmostEqual(degrees[0], 0.0, places=10)

    def test_get_degrees_from_uv_array(self):
        """Test degrees calculation from UV components with array input."""

        xu = np.array([0, 1, 0, -1])
        xv = np.array([1, 0, -1, 0])
        degrees = get_degrees_from_uv(xu, xv)

        # Check specific directions
        expected = np.array([0, 90, 180, 270])
        np.testing.assert_allclose(degrees, expected, rtol=1e-10)

    def test_uv_components_roundtrip(self):
        """Test roundtrip conversion: degrees -> UV -> degrees."""

        original_angles = np.array([0, 45, 90, 135, 180, 225, 270, 315])
        xu, xv = get_uv_components(original_angles)
        recovered_angles = get_degrees_from_uv(xu, xv)

        np.testing.assert_allclose(recovered_angles, original_angles, rtol=1e-10)


class TestCoordinateConversion(unittest.TestCase):
    """Test the coordinate conversion functions."""

    def setUp(self):
        """Set up test data."""

        self.lon = np.array([-3.0, -2.0, -1.0, 0.0, 1.0])
        self.lat = np.array([40.0, 41.0, 42.0, 43.0, 44.0])
        self.projection = "SPAIN"

    def test_convert_lonlat_to_utm_scalar(self):
        """Test longitude/latitude to UTM conversion with scalar input."""

        lon_scalar = np.array([0.0])
        lat_scalar = np.array([40.0])

        utm_x, utm_y = convert_lonlat_to_utm(lon_scalar, lat_scalar, self.projection)

        self.assertEqual(len(utm_x), 1)
        self.assertEqual(len(utm_y), 1)
        self.assertIsInstance(utm_x[0], float)
        self.assertIsInstance(utm_y[0], float)

    def test_convert_lonlat_to_utm_array(self):
        """Test longitude/latitude to UTM conversion with array input."""

        utm_x, utm_y = convert_lonlat_to_utm(self.lon, self.lat, self.projection)

        self.assertEqual(len(utm_x), len(self.lon))
        self.assertEqual(len(utm_y), len(self.lat))
        self.assertTrue(np.all(np.isfinite(utm_x)))
        self.assertTrue(np.all(np.isfinite(utm_y)))

    def test_convert_utm_to_lonlat_scalar(self):
        """Test UTM to longitude/latitude conversion with scalar input."""

        utm_x_scalar = np.array([500000.0])
        utm_y_scalar = np.array([4000000.0])

        lon, lat = convert_utm_to_lonlat(utm_x_scalar, utm_y_scalar, self.projection)

        self.assertEqual(len(lon), 1)
        self.assertEqual(len(lat), 1)
        self.assertIsInstance(lon[0], float)
        self.assertIsInstance(lat[0], float)

    def test_convert_utm_to_lonlat_array(self):
        """Test UTM to longitude/latitude conversion with array input."""

        # First convert to UTM
        utm_x, utm_y = convert_lonlat_to_utm(self.lon, self.lat, self.projection)

        # Then convert back to lon/lat
        lon_back, lat_back = convert_utm_to_lonlat(utm_x, utm_y, self.projection)

        # Check roundtrip conversion
        np.testing.assert_allclose(lon_back, self.lon, rtol=1e-6)
        np.testing.assert_allclose(lat_back, self.lat, rtol=1e-6)

    def test_convert_lonlat_to_utm_meshgrid(self):
        """Test longitude/latitude to UTM conversion with meshgrid input."""

        lon_mesh = np.array([-2.0, -1.0, 0.0])
        lat_mesh = np.array([40.0, 41.0])

        utm_x, utm_y = convert_lonlat_to_utm(lon_mesh, lat_mesh, self.projection)

        # Should return 1D arrays for meshgrid input
        self.assertEqual(len(utm_x), len(lon_mesh))
        self.assertEqual(len(utm_y), len(lat_mesh))

    def test_convert_utm_to_lonlat_meshgrid(self):
        """Test UTM to longitude/latitude conversion with meshgrid input."""

        utm_x_mesh = np.array([500000.0, 600000.0, 700000.0])
        utm_y_mesh = np.array([4000000.0, 4100000.0])

        lon, lat = convert_utm_to_lonlat(utm_x_mesh, utm_y_mesh, self.projection)

        # Should return 1D arrays for meshgrid input
        self.assertEqual(len(lon), len(utm_x_mesh))
        self.assertEqual(len(lat), len(utm_y_mesh))


class TestSpatialGradient(unittest.TestCase):
    """Test the spatial_gradient function."""

    def setUp(self):
        """Set up test data."""

        # Create a simple 2D field with known gradient
        time = pd.date_range("2000-01-01", periods=3)
        lat = np.array([40.0, 41.0, 42.0])
        lon = np.array([-3.0, -2.0, -1.0])

        # Create a linear field: f(x,y) = x + y
        data = np.zeros((3, 3, 3))
        for t in range(3):
            for i, la in enumerate(lat):
                for j, lo in enumerate(lon):
                    data[t, i, j] = la + lo

        self.data = xr.DataArray(
            data,
            coords={"time": time, "latitude": lat, "longitude": lon},
            dims=["time", "latitude", "longitude"],
        )

    def test_spatial_gradient_basic(self):
        """Test basic spatial gradient calculation."""

        gradient = spatial_gradient(self.data)

        # Check output structure
        self.assertEqual(gradient.dims, self.data.dims)
        self.assertEqual(gradient.shape, self.data.shape)
        self.assertIn("units", gradient.attrs)
        self.assertIn("name", gradient.attrs)

        # Check that gradient is finite
        self.assertTrue(np.all(np.isfinite(gradient.values)))

    def test_spatial_gradient_attributes(self):
        """Test that gradient has correct attributes."""

        gradient = spatial_gradient(self.data)

        self.assertEqual(gradient.attrs["units"], "m^2/s^2")
        self.assertEqual(gradient.attrs["name"], "Gradient")

    def test_spatial_gradient_non_zero(self):
        """Test that gradient is non-zero for non-constant field."""

        gradient = spatial_gradient(self.data)

        # For a linear field, gradient should be non-zero
        self.assertGreater(np.abs(gradient.values).max(), 0.0)


class TestAngleConversion(unittest.TestCase):
    """Test the angle conversion functions."""

    def test_nautical_to_mathematical_scalar(self):
        """Test nautical to mathematical conversion with scalar input."""

        # Test key directions
        self.assertAlmostEqual(
            nautical_to_mathematical(0), 90, places=10
        )  # North -> East
        self.assertAlmostEqual(
            nautical_to_mathematical(90), 0, places=10
        )  # East -> East
        self.assertAlmostEqual(
            nautical_to_mathematical(180), 270, places=10
        )  # South -> West
        self.assertAlmostEqual(
            nautical_to_mathematical(270), 180, places=10
        )  # West -> South

    def test_nautical_to_mathematical_array(self):
        """Test nautical to mathematical conversion with array input."""

        nautical_angles = np.array([0, 90, 180, 270])
        mathematical_angles = nautical_to_mathematical(nautical_angles)

        expected = np.array([90, 0, 270, 180])
        np.testing.assert_allclose(mathematical_angles, expected, rtol=1e-10)

    def test_mathematical_to_nautical_scalar(self):
        """Test mathematical to nautical conversion with scalar input."""

        # Test key directions
        self.assertAlmostEqual(
            mathematical_to_nautical(90), 0, places=10
        )  # East -> North
        self.assertAlmostEqual(
            mathematical_to_nautical(0), 90, places=10
        )  # East -> East
        self.assertAlmostEqual(
            mathematical_to_nautical(270), 180, places=10
        )  # West -> South
        self.assertAlmostEqual(
            mathematical_to_nautical(180), 270, places=10
        )  # South -> West

    def test_mathematical_to_nautical_array(self):
        """Test mathematical to nautical conversion with array input."""

        mathematical_angles = np.array([90, 0, 270, 180])
        nautical_angles = mathematical_to_nautical(mathematical_angles)

        expected = np.array([0, 90, 180, 270])
        np.testing.assert_allclose(nautical_angles, expected, rtol=1e-10)

    def test_angle_conversion_roundtrip(self):
        """Test roundtrip conversion: nautical -> mathematical -> nautical."""

        original_nautical = np.array([0, 45, 90, 135, 180, 225, 270, 315])
        mathematical = nautical_to_mathematical(original_nautical)
        recovered_nautical = mathematical_to_nautical(mathematical)

        np.testing.assert_allclose(recovered_nautical, original_nautical, rtol=1e-10)

    def test_mathematical_to_nautical_zero_case(self):
        """Test mathematical to nautical conversion with zero input."""

        result = mathematical_to_nautical(0)
        self.assertAlmostEqual(result, 90, places=10)


if __name__ == "__main__":
    unittest.main()



from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.linear_model import LinearRegression

from ..core.decorators import validate_data_kma
from ._base_datamining import BaseClustering


class KMAError(Exception):
    """
    Custom exception for KMA class.
    """

    def __init__(self, message: str = "KMA error occurred."):
        self.message = message
        super().__init__(self.message)


class KMA(BaseClustering):
    """
    K-Means Algorithm (KMA) class.

    This class performs the K-Means algorithm on a given dataframe.

    Attributes
    ----------
    num_clusters : int
        The number of clusters to use in the K-Means algorithm.
    seed : int
        The random seed to use as initial datapoint.
    data_variables : List[str]
        A list with all data variables.
    directional_variables : List[str]
        A list with directional variables.
    fitting_variables : List[str]
        A list with fitting variables.
    custom_scale_factor : dict
        A dictionary of custom scale factors.
    scale_factor : dict
        A dictionary of scale factors (after normalizing the data).
    centroids : pd.DataFrame
        The selected centroids.
    normalized_centroids : pd.DataFrame
        The selected normalized centroids.
    centroid_real_indices : np.array
        The real indices of the selected centroids.
    is_fitted : bool
        A flag indicating whether the model is fitted or not.

    Examples
    --------
    .. jupyter-execute::

        import numpy as np
        import pandas as pd
        from bluemath_tk.datamining.kma import KMA

        data = pd.DataFrame(
            {
                "Hs": np.random.rand(1000) * 7,
                "Tp": np.random.rand(1000) * 20,
                "Dir": np.random.rand(1000) * 360
            }
        )
        kma = KMA(num_clusters=5)
        nearest_centroids_idxs, nearest_centroids_df = kma.fit_predict(
            data=data,
            directional_variables=["Dir"],
        )

        kma.plot_selected_centroids(plot_text=True)
    """

    def __init__(
        self,
        num_clusters: int,
        seed: int = None,
    ) -> None:
        """
        Initializes the KMA class.

        Parameters
        ----------
        num_clusters : int
            The number of clusters to use in the K-Means algorithm.
            Must be greater than 0.
        seed : int, optional
            The random seed to use as initial datapoint.
            Must be greater or equal to 0 and less than number of datapoints.
            Default is 0.

        Raises
        ------
        ValueError
            If num_clusters is not greater than 0.
            Or if seed is not greater or equal to 0.
        """

        super().__init__()
        self.set_logger_name(name=self.__class__.__name__)

        if num_clusters > 0:
            self.num_clusters = int(num_clusters)
        else:
            raise ValueError("Variable num_clusters must be > 0")
        if seed is None:
            self.seed = None
        elif seed >= 0:
            self.seed = int(seed)
        else:
            raise ValueError("Variable seed must be >= 0")
        self._kma = KMeans(
            n_clusters=self.num_clusters,
            random_state=self.seed,
        )
        self.logger.info(
            f"KMA object created with {self.num_clusters} clusters and seed {self.seed}."
            "To customize kma, do self.kma = dict(n_clusters=..., random_state=..., etc)"
        )

        self._data: pd.DataFrame = pd.DataFrame()
        self._normalized_data: pd.DataFrame = pd.DataFrame()
        self._data_to_fit: pd.DataFrame = pd.DataFrame()
        self.data_variables: List[str] = []
        self.directional_variables: List[str] = []
        self.fitting_variables: List[str] = []
        self.custom_scale_factor: dict = {}
        self.scale_factor: dict = {}
        self.centroids: pd.DataFrame = pd.DataFrame()
        self.normalized_centroids: pd.DataFrame = pd.DataFrame()
        self.centroid_real_indices: np.array = np.array([])
        self.is_fitted: bool = False
        self.regression_guided: dict = {}

    @property
    def kma(self) -> KMeans:
        return self._kma

    @kma.setter
    def kma(self, kma_params_dict) -> None:
        """
        Setter for the KMeans object.

        Parameters
        ----------
        kma_params_dict : dict
            A dictionary with KMeans parameters.
            The keys should be the same as the KMeans parameters.
            Example: {"n_clusters": 5, "random_state": 42}
        """

        self._kma = KMeans(**kma_params_dict)

    @property
    def data(self) -> pd.DataFrame:
        """
        Returns the original data used for clustering.
        """

        return self._data

    @property
    def normalized_data(self) -> pd.DataFrame:
        """
        Returns the normalized data used for clustering.
        """

        return self._normalized_data

    @property
    def data_to_fit(self) -> pd.DataFrame:
        """
        Returns the data used for fitting the K-Means algorithm.
        """

        return self._data_to_fit

    @staticmethod
    def add_regression_guided(
        data: pd.DataFrame, vars: List[str], alpha: List[float]
    ) -> pd.DataFrame:
        """
        Calculate regression-guided variables.

        Parameters
        ----------
        data : pd.DataFrame
            The data to fit the K-Means algorithm.
        vars : List[str]
            The variables to use for regression-guided clustering.
        alpha : List[float]
            The alpha values to use for regression-guided clustering.

        Returns
        -------
        pd.DataFrame
            The data with the regression-guided variables.
        """

        # Stack guiding variables into (time, n_vars) array
        X = data.drop(columns=vars)
        Y = np.stack([data[var].values for var in vars], axis=1)

        # Normalize input features
        X_std = X.std().replace(0, 1)
        X_norm = X / X_std

        # Add intercept column to input
        X_design = np.column_stack((np.ones(len(X)), X_norm.values))

        # Normalize guiding targets
        Y_std = np.nanstd(Y, axis=0)
        Y_std[Y_std == 0] = 1.0

        # Fit regression model to predict guiding vars from input
        model = LinearRegression(fit_intercept=False).fit(X_design, Y / Y_std)
        Y_pred = model.predict(X_design) * Y_std  # De-normalize predictions

        # Weight columns by input alpha
        X_weight = 1.0 - np.sum(alpha)
        X_scaled = X_weight * X.values
        Y_scaled = Y_pred * alpha

        df = pd.DataFrame(np.hstack([X_scaled, Y_scaled]), index=data.index)
        df.columns = list(X.columns) + vars

        return df

    @validate_data_kma
    def fit(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        min_number_of_points: int = None,
        max_number_of_iterations: int = 10,
        normalize_data: bool = False,
        regression_guided: Dict[str, List] = {},
    ) -> None:
        """
        Fit the K-Means algorithm to the provided data.

        TODO: Add option to force KMA initialization with MDA centroids.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the KMA algorithm.
        directional_variables : List[str], optional
            A list of directional variables that will be transformed to u and v components.
            Then, to use custom_scale_factor, you must specify the variables names with the u and v suffixes.
            Example: directional_variables=["Dir"], custom_scale_factor={"Dir_u": [0, 1], "Dir_v": [0, 1]}.
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            If normalize_data is True, this will be used to normalize the data.
            Example: {"Hs": [0, 10], "Tp": [0, 10]}.
            Default is {}.
        min_number_of_points : int, optional
            The minimum number of points to consider a cluster.
            Default is None.
        max_number_of_iterations : int, optional
            The maximum number of iterations for the K-Means algorithm.
            This is used when min_number_of_points is not None.
            Default is 10.
        normalize_data : bool, optional
            A flag to normalize the data.
            If True, the data will be normalized using the custom_scale_factor.
            Default is False.
        regression_guided: dict, optional
            A dictionary specifying regression-guided clustering variables and relative weights.
            Example: {"vars": ["Fe"], "alpha": [0.6]}.
            Default is {}.
        """

        if regression_guided:
            data = self.add_regression_guided(
                data=data,
                vars=regression_guided.get("vars", None),
                alpha=regression_guided.get("alpha", None),
            )

        super().fit(
            data=data,
            directional_variables=directional_variables,
            custom_scale_factor=custom_scale_factor,
            normalize_data=normalize_data,
        )

        # Fit K-Means algorithm
        if min_number_of_points is not None:
            stable_kma_child = False
            number_of_tries = 0
            while not stable_kma_child:
                kma_child = KMeans(n_clusters=self.num_clusters)
                predicted_labels = kma_child.fit_predict(self.normalized_data)
                _unique_labels, counts = np.unique(predicted_labels, return_counts=True)
                if np.all(counts >= min_number_of_points):
                    stable_kma_child = True
                number_of_tries += 1
                if number_of_tries > max_number_of_iterations:
                    raise ValueError(
                        f"Failed to find a stable K-Means configuration after {max_number_of_iterations} attempts."
                        "Change max_number_of_iterations or min_number_of_points."
                    )
            self.logger.info(
                f"Found a stable K-Means configuration after {number_of_tries} attempts."
            )
            self._kma = kma_child
        else:
            self._kma = self.kma.fit(self.normalized_data)

        # Calculate the centroids
        self.centroid_real_indices = self.kma.labels_.copy()
        self.normalized_centroids = pd.DataFrame(
            self.kma.cluster_centers_, columns=self.fitting_variables
        )
        self.centroids = self.denormalize(
            normalized_data=self.normalized_centroids, scale_factor=self.scale_factor
        )

        for directional_variable in self.directional_variables:
            self.centroids[directional_variable] = self.get_degrees_from_uv(
                xu=self.centroids[f"{directional_variable}_u"].values,
                xv=self.centroids[f"{directional_variable}_v"].values,
            )

        # Set the fitted flag to True
        self.is_fitted = True

    def predict(self, data: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Predict the nearest centroid for the provided data.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the prediction.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point,
            and the nearest centroids.
        """

        if self.is_fitted is False:
            raise KMAError("KMA model is not fitted.")

        normalized_data = super().predict(data=data)

        y = self.kma.predict(X=normalized_data)

        return pd.DataFrame(
            y, columns=["kma_bmus"], index=data.index
        ), self.centroids.iloc[y]

    def fit_predict(
        self,
        data: pd.DataFrame,
        directional_variables: List[str] = [],
        custom_scale_factor: dict = {},
        min_number_of_points: int = None,
        max_number_of_iterations: int = 10,
        normalize_data: bool = False,
        regression_guided: Dict[str, List] = {},
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Fit the K-Means algorithm to the provided data and predict the nearest centroid
        for each data point.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be used for the KMA algorithm.
        directional_variables : List[str], optional
            A list of directional variables that will be transformed to u and v components.
            Then, to use custom_scale_factor, you must specify the variables names with the u and v suffixes.
            Example: directional_variables=["Dir"], custom_scale_factor={"Dir_u": [0, 1], "Dir_v": [0, 1]}.
            Default is [].
        custom_scale_factor : dict, optional
            A dictionary specifying custom scale factors for normalization.
            If normalize_data is True, this will be used to normalize the data.
            Example: {"Hs": [0, 10], "Tp": [0, 10]}.
            Default is {}.
        min_number_of_points : int, optional
            The minimum number of points to consider a cluster.
            Default is None.
        max_number_of_iterations : int, optional
            The maximum number of iterations for the K-Means algorithm.
            This is used when min_number_of_points is not None.
            Default is 10.
        normalize_data : bool, optional
            A flag to normalize the data.
            If True, the data will be normalized using the custom_scale_factor.
            Default is False.
        regression_guided: dict, optional
            A dictionary specifying regression-guided clustering variables and relative weights.
            Example: {"vars": ["Fe"], "alpha": [0.6]}.
            Default is {}.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            A tuple containing the nearest centroid index for each data point,
            and the nearest centroids.
        """

        self.fit(
            data=data,
            directional_variables=directional_variables,
            custom_scale_factor=custom_scale_factor,
            min_number_of_points=min_number_of_points,
            max_number_of_iterations=max_number_of_iterations,
            normalize_data=normalize_data,
            regression_guided=regression_guided,
        )

        return self.predict(data=data)



import calendar
import json
import os
from typing import List

import cdsapi
import xarray as xr

from .._base_downloaders import BaseDownloader

config = {
    "url": "https://cds.climate.copernicus.eu/api",  # /v2?
    "key": "your-api-token",
}


class CopernicusDownloader(BaseDownloader):
    """
    This is the main class to download data from the Copernicus Climate Data Store.

    Attributes
    ----------
    product : str
        The product to download data from. Currently only ERA5 is supported.
    product_config : dict
        The configuration for the product to download data from.
    client : cdsapi.Client
        The client to interact with the Copernicus Climate Data Store API.

    Examples
    --------
    .. jupyter-execute::

        from bluemath_tk.downloaders.copernicus.copernicus_downloader import CopernicusDownloader

        copernicus_downloader = CopernicusDownloader(
            product="ERA5",
            base_path_to_download="/path/to/Copernicus/",  # Will be created if not available
            token=None,
            check=True,
        )
        result = copernicus_downloader.download_data_era5(
            variables=["swh"],
            years=["2020"],
            months=["01", "03"],
        )
        print(result)
    """

    products_configs = {
        "ERA5": json.load(
            open(os.path.join(os.path.dirname(__file__), "ERA5", "ERA5_config.json"))
        )
    }

    def __init__(
        self,
        product: str,
        base_path_to_download: str,
        token: str = None,
        debug: bool = True,
        check: bool = True,
    ) -> None:
        """
        This is the constructor for the CopernicusDownloader class.

        Parameters
        ----------
        product : str
            The product to download data from. Currently only ERA5 is supported.
        base_path_to_download : str
            The base path to download the data to.
        token : str, optional
            The API token to use to download data. Default is None.
        debug : bool, optional
            Whether to run in debug mode. Default is True.
        check : bool, optional
            Whether to just check the data. Default is True.

        Raises
        ------
        ValueError
            If the product configuration is not found.
        """

        super().__init__(
            base_path_to_download=base_path_to_download, debug=debug, check=check
        )
        self._product = product
        self._product_config = self.products_configs.get(product)
        if self._product_config is None:
            raise ValueError(f"{product} configuration not found")
        self.set_logger_name(
            f"CopernicusDownloader-{product}", level="DEBUG" if debug else "INFO"
        )
        if not self.check:
            self._client = cdsapi.Client(
                url=config["url"], key=token or config["key"], debug=self.debug
            )
            self.logger.info("---- DOWNLOADING DATA ----")
        else:
            self.logger.info("---- CHECKING DATA ----")

    @property
    def product(self) -> str:
        return self._product

    @property
    def product_config(self) -> dict:
        return self._product_config

    @property
    def client(self) -> cdsapi.Client:
        return self._client

    def list_variables(self, type: str = None) -> List[str]:
        """
        Lists the variables available for the product.
        Filtering by type if provided.

        Parameters
        ----------
        type : str, optional
            The type of variables to list. Default is None.

        Returns
        -------
        List[str]
            The list of variables available for the product.
        """

        if type == "ocean":
            return [
                var_name
                for var_name, var_info in self.product_config["variables"].items()
                if var_info["type"] == "ocean"
            ]
        return list(self.product_config["variables"].keys())

    def list_datasets(self) -> List[str]:
        """
        Lists the datasets available for the product.

        Returns
        -------
        List[str]
            The list of datasets available for the product.
        """

        return list(self.product_config["datasets"].keys())

    def show_markdown_table(self) -> None:
        """
        Create a Markdown table from the configuration dictionary and print it.
        """

        # Define the table headers
        headers = ["name", "long_name", "units", "type"]
        header_line = "| " + " | ".join(headers) + " |"
        separator_line = (
            "| " + " | ".join(["-" * len(header) for header in headers]) + " |"
        )

        # Initialize the table with headers
        table_lines = [header_line, separator_line]

        # Add rows for each variable
        for var_name, var_info in self.product_config["variables"].items():
            long_name = var_info.get("long_name", "")
            units = var_info.get("units", "")
            type = var_info.get("type", "")
            row = f"| {var_name} | {long_name} | {units} | {type} |"
            table_lines.append(row)

        # Print the table
        print("\n".join(table_lines))

    def download_data(self, *args, **kwargs) -> str:
        """
        Downloads the data for the product.

        Parameters
        ----------
        *args
            The arguments to pass to the download function.
        **kwargs
            The keyword arguments to pass to the download function.

        Returns
        -------
        str
            The message with the fully downloaded files and the not fully downloaded files.

        Raises
        ------
        ValueError
            If the product is not supported.
        """

        if self.product == "ERA5":
            return self.download_data_era5(*args, **kwargs)
        else:
            raise ValueError(f"Download for product {self.product} not supported")

    def download_data_era5(
        self,
        variables: List[str],
        years: List[str],
        months: List[str],
        days: List[str] = None,
        times: List[str] = None,
        area: List[float] = None,
        product_type: str = "reanalysis",
        data_format: str = "netcdf",
        download_format: str = "unarchived",
        force: bool = False,
    ) -> str:
        """
        Downloads the data for the ERA5 product.

        Parameters
        ----------
        variables : List[str]
            The variables to download. If not provided, all variables in self.product_config
            will be downloaded.
        years : List[str]
            The years to download. Years are downloaded one by one.
        months : List[str]
            The months to download. Months are downloaded together.
        days : List[str], optional
            The days to download. If None, all days in the month will be downloaded.
            Default is None.
        times : List[str], optional
            The times to download. If None, all times in the day will be downloaded.
            Default is None.
        area : List[float], optional
            The area to download. If None, the whole globe will be downloaded.
            Default is None.
        product_type : str, optional
            The product type to download. Default is "reanalysis".
        data_format : str, optional
            The data format to download. Default is "netcdf".
        download_format : str, optional
            The download format to use. Default is "unarchived".
        force : bool, optional
            Whether to force the download. Default is False.

        Returns
        -------
        str
            The message with the fully downloaded files and the not fully downloaded files.
            Error files are also included.

        TODO
        -----
        - Implement lambda function to name the files.
        """

        if not isinstance(variables, list):
            raise ValueError("Variables must be a list of strings")
        elif len(variables) == 0:
            variables = list(self.product_config["variables"].keys())
            self.logger.info(f"Variables not provided. Using {variables}")
        if not isinstance(years, list) or len(years) == 0:
            raise ValueError("Years must be a non-empty list of strings")
        else:
            years = [f"{int(year):04d}" for year in years]
        if not isinstance(months, list) or len(months) == 0:
            raise ValueError("Months must be a non-empty list of strings")
        else:
            months = [f"{int(month):02d}" for month in months]
            last_month = months[-1]
        if days is not None:
            if not isinstance(days, list) or len(days) == 0:
                raise ValueError("Day must be a non-empty list of strings")
        else:
            days = [f"{day:02d}" for day in range(1, 32)]
            self.logger.info(f"Day not provided. Using {days}")
        if times is not None:
            if not isinstance(times, list) or len(times) == 0:
                raise ValueError("Time must be a non-empty list of strings")
        else:
            times = [f"{hour:02d}:00" for hour in range(24)]
            self.logger.info(f"Time not provided. Using {times}")
        if area is not None:
            if not isinstance(area, list) or len(area) != 4:
                raise ValueError("Area must be a list of 4 floats")
        if not isinstance(product_type, str):
            raise ValueError("Product type must be a string")
        if not isinstance(data_format, str):
            raise ValueError("Data format must be a string")
        if not isinstance(download_format, str):
            raise ValueError("Download format must be a string")
        if not isinstance(force, bool):
            raise ValueError("Force must be a boolean")

        fully_downloaded_files: List[str] = []
        NOT_fullly_downloaded_files: List[str] = []
        error_files: List[str] = []

        for variable in variables:
            for year in years:
                variable_config = self.product_config["variables"].get(variable)
                if variable_config is None:
                    self.logger.error(
                        f"Variable {variable} not found in product configuration file"
                    )
                    continue
                variable_dataset = self.product_config["datasets"].get(
                    variable_config["dataset"]
                )
                if variable_dataset is None:
                    self.logger.error(
                        f"Dataset {variable_config['dataset']} not found in product configuration file"
                    )
                    continue

                template_for_variable = variable_dataset["template"].copy()
                if variable == "spectra":
                    template_for_variable["date"] = (
                        f"{year}-{months[0]}-01/to/{year}-{months[-1]}-31"
                    )
                    if area is not None:
                        template_for_variable["area"] = "/".join(
                            [str(coord) for coord in area]
                        )
                else:
                    template_for_variable["variable"] = variable_config["cds_name"]
                    template_for_variable["year"] = year
                    template_for_variable["month"] = months
                    template_for_variable["day"] = days
                    template_for_variable["time"] = times
                    template_for_variable["product_type"] = product_type
                    template_for_variable["data_format"] = data_format
                    template_for_variable["download_format"] = download_format
                    if area is not None:
                        template_for_variable["area"] = area

                self.logger.info(
                    f"""
                    Template for variable {variable}:
                    {template_for_variable}
                    """
                )

                skip_because_of_manadatory_fields = False
                for mandatory_field in variable_dataset["mandatory_fields"]:
                    try:
                        if template_for_variable.get(mandatory_field) is None:
                            template_for_variable[mandatory_field] = variable_config[
                                mandatory_field
                            ]
                    except KeyError:
                        self.logger.error(
                            f"Mandotory field {mandatory_field} not found in variable configuration file for {variable}"
                        )
                        skip_because_of_manadatory_fields = True
                if skip_because_of_manadatory_fields:
                    continue

                # Create the output file name once request is properly formatted
                output_nc_file = os.path.join(
                    self.base_path_to_download,
                    self.product,
                    variable_config["dataset"],
                    variable_config["type"],
                    product_type,
                    variable_config["cds_name"],
                    f"{variable_config['nc_name']}_{year}_{'_'.join(months)}.nc",
                    # f"era5_waves_{variable_config['cds_name']}_{year}.nc",
                )
                # Create the output directory if it does not exist
                if not self.check:
                    os.makedirs(os.path.dirname(output_nc_file), exist_ok=True)

                self.logger.info(f"""
                                 
                    Analyzing {output_nc_file}

                """)

                try:
                    if self.check or not force:
                        if os.path.exists(output_nc_file):
                            self.logger.debug(
                                f"Checking {output_nc_file} file is complete"
                            )
                            try:
                                nc = xr.open_dataset(output_nc_file)
                                _, last_day = calendar.monthrange(
                                    int(year), int(last_month)
                                )
                                last_hour = f"{year}-{last_month}-{last_day}T23"
                                try:
                                    last_hour_nc = str(nc.time[-1].values)
                                except Exception as _te:
                                    last_hour_nc = str(nc.valid_time[-1].values)
                                nc.close()
                                if last_hour not in last_hour_nc:
                                    self.logger.debug(
                                        f"{output_nc_file} ends at {last_hour_nc} instead of {last_hour}"
                                    )
                                    if self.check:
                                        NOT_fullly_downloaded_files.append(
                                            output_nc_file
                                        )
                                    else:
                                        self.logger.debug(
                                            f"Downloading: {variable} to {output_nc_file} because it is not complete"
                                        )
                                        self.client.retrieve(
                                            name=variable_config["dataset"],
                                            request=template_for_variable,
                                            target=output_nc_file,
                                        )
                                        fully_downloaded_files.append(output_nc_file)
                                else:
                                    self.logger.debug(
                                        f"{output_nc_file} already downloaded and complete"
                                    )
                                    fully_downloaded_files.append(output_nc_file)
                            except Exception as e:
                                self.logger.error(
                                    f"Error was raised opening {output_nc_file} - {e}, re-downloading..."
                                )
                                if self.check:
                                    NOT_fullly_downloaded_files.append(output_nc_file)
                                else:
                                    self.logger.debug(
                                        f"Downloading: {variable} to {output_nc_file} because it is not complete"
                                    )
                                    self.client.retrieve(
                                        name=variable_config["dataset"],
                                        request=template_for_variable,
                                        target=output_nc_file,
                                    )
                                    fully_downloaded_files.append(output_nc_file)
                        elif self.check:
                            NOT_fullly_downloaded_files.append(output_nc_file)
                        else:
                            self.logger.debug(
                                f"Downloading: {variable} to {output_nc_file}"
                            )
                            self.client.retrieve(
                                name=variable_config["dataset"],
                                request=template_for_variable,
                                target=output_nc_file,
                            )
                            fully_downloaded_files.append(output_nc_file)
                    else:
                        self.logger.debug(
                            f"Downloading: {variable} to {output_nc_file}"
                        )
                        self.client.retrieve(
                            name=variable_config["dataset"],
                            request=template_for_variable,
                            target=output_nc_file,
                        )
                        fully_downloaded_files.append(output_nc_file)

                except Exception as e:
                    self.logger.error(f"""
                                        
                        Skippping {output_nc_file} for {e}

                    """)
                    error_files.append(output_nc_file)

        fully_downloaded_files_str = "\n".join(fully_downloaded_files)
        NOT_fullly_downloaded_files_str = "\n".join(NOT_fullly_downloaded_files)
        error_files = "\n".join(error_files)

        return f"""
            Fully downloaded files:
            {fully_downloaded_files_str}
            Not fully downloaded files:
            {NOT_fullly_downloaded_files_str}
            Error files:
            {error_files}
        """



from typing import Dict, Tuple

import xarray as xr


def superpoint_calculation(
    stations_data: xr.DataArray,
    stations_dimension_name: str,
    sectors_for_each_station: Dict[str, Tuple[float, float]],
    overlap_angle: float = 0.0,
) -> xr.DataArray:
    """
    Join multiple station spectral data for each directional sector.

    Parameters
    ----------
    stations_data : xr.DataArray
        DataArray containing spectral data for multiple stations.
    stations_dimension_name : str
        Name of the dimension representing different stations in the DataArray.
    sectors_for_each_station : Dict[str, Tuple[float, float]]
        Dictionary mapping each station ID to a tuple of (min_direction, max_direction)
        representing the directional sector for that station.
    overlap_angle : float
        Angle in degrees that defines the overlap between two sectors.

    Returns
    -------
    xr.DataArray
        A new DataArray where each point is the sum of spectral data from all stations
        for the specified directional sector.

    Notes
    -----
    If your stations_data is saved in different files, you can load them all and then
    concatenate them using xr.open_mfdataset function. Example below:

    ```python
    files = [
        "path/to/station1.nc",
        "path/to/station2.nc",
        "path/to/station3.nc"
    ]

    def load_station_data(ds: xr.Dataset) -> xr.DataArray:
        return ds.efth.expand_dims("station")

    stations_data = xr.open_mfdataset(
        files,
        concat_dim="station",
        preprocess=load_station_data,
    )
    ```
    """

    superpoint_dataarray = xr.zeros_like(
        stations_data.isel({stations_dimension_name: 0})
    )

    if overlap_angle == 0:
        for station_id, (dir_min, dir_max) in sectors_for_each_station.items():
            if dir_min < dir_max:
                mask = (stations_data["dir"] >= dir_min) & (
                    stations_data["dir"] < dir_max
                )
            else:
                # Handle wrap-around (e.g., 350° to 10°)
                mask = (stations_data["dir"] >= dir_min) | (
                    stations_data["dir"] < dir_max
                )
            superpoint_dataarray += stations_data.sel(
                {stations_dimension_name: station_id}
            ).where(mask, 0.0)

    else:
        # With overlap - expand sectors and average overlaps
        directions = stations_data["dir"]
        count_array = xr.zeros_like(superpoint_dataarray)  # Counter for overlaps

        for station_id, (dir_min, dir_max) in sectors_for_each_station.items():
            station_data = stations_data.sel({stations_dimension_name: station_id})

            # Expand sector boundaries by overlap_angle
            if (dir_max - dir_min) < 0:
                mask = (directions >= dir_min - overlap_angle) | (
                    directions <= dir_max + overlap_angle
                )
            else:
                mask = (directions >= dir_min - overlap_angle) & (
                    directions <= dir_max + overlap_angle
                )

            # Add contribution where mask is true
            superpoint_dataarray += station_data.where(mask, 0.0)
            count_array += xr.where(mask, 1, 0)

        # Average where there are overlaps (count > 1)
        overlap_mask = count_array > 1
        superpoint_dataarray = xr.where(
            overlap_mask, superpoint_dataarray / count_array, superpoint_dataarray
        )

    return superpoint_dataarray



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta, date
from dateutil.relativedelta import relativedelta
from cftime._cftime import DatetimeGregorian
import calendar

import numpy as np
import pandas as pd
import xarray as xr

# TODO replace xds2datetime with date2datenum (needs new date type switch) 
# TODO make a fast_reindex daily option
# TODO optimize, refactor, and tests


# MATLAB TIMES

def datematlab2datetime(datenum_matlab):
    'Return python datetime for matlab datenum. Transform and adjust from matlab.'

    d = datetime.fromordinal(int(datenum_matlab)) + \
    timedelta(days=float(datenum_matlab % 1)) - \
        timedelta(days=366) + timedelta(microseconds=0)

    return d

def datevec2datetime(d_vec):
    '''
    Returns datetime list from a datevec matrix
    d_vec = [[y1 m1 d1 H1 M1],[y2 ,2 d2 H2 M2],..]
    '''

    return [datetime(d[0], d[1], d[2], d[3], d[4]) for d in d_vec]

def DateConverter_Mat2Py(datearray_matlab):
    'Parses matlab datenum array to python datetime list'

    return [datematlab2datetime(x) for x in datearray_matlab]


# PYTHON TIMES

def xds2datetime(d64):
    'converts xr.Dataset np.datetime64[ns] into datetime'
    # TODO: hour minutes and seconds 

    return datetime(d64.dt.year, d64.dt.month, d64.dt.day)

def npdt64todatetime(dt64):
    'converts np.datetime64[ns] into datetime'

    ts = (dt64 - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
    return datetime(1970, 1, 1) + timedelta(seconds=ts)

# declare aux function for generating datetime.datetime array
def generate_datetimes(t0, t1, dtype='datetime64[h]'):

    # lambda vectorized: datetime.datetime from utc timestamp
    lb_dfts = lambda t: datetime.utcfromtimestamp(t)
    np_dfts = np.vectorize(lb_dfts)

    # mount times array
    dtd = {
        'datetime64[h]': timedelta(hours=1),
        'datetime64[D]': timedelta(days=1),
    }
    tdd = dtd[dtype]

    tg = np.arange(t0, t1 + tdd, dtype=dtype)
    tg = (tg - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
    tg = np_dfts(tg)

    return tg

def fast_reindex_hourly(xds_data):
    '''
    Fast and secure method to reindex (pad) xarray.Dataset to hourly data

    xds_data - xarray.Dataset with time coordinate
    '''

    # def aux function for getting timedeltas as int array
    def get_deltas(t):
        t_df = (np.diff(t))
        t_tp = type(t_df[0])

        # total number of hours each date interval
        if  t_tp == np.timedelta64:
            d = t_df.astype('timedelta64[h]') / np.timedelta64(1,'h')
            d = d.astype(int)

        else:
            # lambda vectorized: datetime.timedelta to number of hours 
            lb_ghs = lambda t: t.days*24
            np_ghs = np.vectorize(lb_ghs)
            d = np_ghs(t_df)

        return d

    # generate output time array
    time_base = xds_data.time.values[:]
    t0, t1 = date2datenum(time_base[0]), date2datenum(time_base[-1])
    time_h = generate_datetimes(t0, t1, dtype='datetime64[h]')

    # repeat data in each variable new lower time frame (pad) 
    ds = get_deltas(time_base)
    dv = {}
    for vn in xds_data.keys():
        b = xds_data[vn].values[:]
        p = np.append(np.repeat(b[:-1], ds), b[-1])
        dv[vn] = (('time',), p)

    # return xarray.Dataset
    xds_out = xr.Dataset(dv, coords={'time': time_h})

    return xds_out

def fast_reindex_hourly_nsim(xds_data):
    '''
    Fast and secure method to reindex (pad) xarray.Dataset to hourly data

    xds_data - xarray.Dataset with time, n_sim coordinates
    '''

    # def aux function for getting timedeltas as int array
    def get_deltas(t):
        t_df = (np.diff(t))
        t_tp = type(t_df[0])

        # total number of hours each date interval
        if  t_tp == np.timedelta64:
            d = t_df.astype('timedelta64[h]') / np.timedelta64(1,'h')
            d = d.astype(int)

        else:
            # lambda vectorized: datetime.timedelta to number of hours 
            lb_ghs = lambda t: t.days*24
            np_ghs = np.vectorize(lb_ghs)
            d = np_ghs(t_df)

        return d

    # generate output time array
    time_base = xds_data.time.values[:]
    t0, t1 = date2datenum(time_base[0]), date2datenum(time_base[-1])
    time_h = generate_datetimes(t0, t1, dtype='datetime64[h]')

    # repeat data in each variable new lower time frame (pad) 
    ds = get_deltas(time_base)

    # iterate n_sim dimension
    l_sim = []
    for s in xds_data.n_sim:
        xds_d = xds_data.sel(n_sim=s)

        # fast reindex
        dv = {}
        for vn in xds_d.keys():
            b = xds_d[vn].values[:]
            p = np.append(np.repeat(b[:-1], ds), b[-1])
            dv[vn] = (('time',), p)
        # return xarray.Dataset
        xds_rd = xr.Dataset(dv, coords={'time': time_h})
        l_sim.append(xds_rd)

    # concat simulations
    xds_out = xr.concat(l_sim, 'n_sim')

    return xds_out

def xds_reindex_daily(xds_data,  dt_lim1=None, dt_lim2=None):
    '''
    Reindex xarray.Dataset to daily data between optional limits
    '''

    # TODO: remove limits from inside function

    # TODO: remove this swich and use date2datenum
    if isinstance(xds_data.time.values[0], datetime):
        xds_dt1 = xds_data.time.values[0]
        xds_dt2 = xds_data.time.values[-1]
    else:
        # parse xds times to python datetime
        xds_dt1 = xds2datetime(xds_data.time[0])
        xds_dt2 = xds2datetime(xds_data.time[-1])

    # cut data at limits
    if dt_lim1:
        xds_dt1 = max(xds_dt1, dt_lim1)
    if dt_lim2:
        xds_dt2 = min(xds_dt2, dt_lim2)

    # number of days
    num_days = (xds_dt2-xds_dt1).days+1

    # reindex xarray.Dataset
    return xds_data.reindex(
        {'time': [xds_dt1 + timedelta(days=i) for i in range(num_days)]},
        method = 'pad',
    )

def xds_reindex_monthly(xds_data):
    '''
    Reindex xarray.Dataset to monthly data
    '''

    # TODO: remove this swich and use date2datenum
    if isinstance(xds_data.time.values[0], datetime):
        xds_dt1 = xds_data.time.values[0]
        xds_dt2 = xds_data.time.values[-1]
    else:
        # parse xds times to python datetime
        xds_dt1 = xds2datetime(xds_data.time[0])
        xds_dt2 = xds2datetime(xds_data.time[-1])

    # number of months
    num_months = (xds_dt2.year - xds_dt1.year)*12 + \
            xds_dt2.month - xds_dt1.month

    # reindex xarray.Dataset
    return xds_data.reindex(
        {'time': [xds_dt1 + relativedelta(months=i) for i in range(num_months)]},
        method = 'pad',
    )

def xds_common_dates_daily(xds_list):
    '''
    returns daily datetime array between a list of xarray.Dataset comon date
    limits
    '''

    d1, d2 = xds_limit_dates(xds_list)

    return [d1 + timedelta(days=i) for i in range((d2-d1).days+1)]

def xds_limit_dates(xds_list):
    '''
    returns datetime common limits between a list of xarray.Dataset
    '''

    d1 = None
    d2 = None

    for xds_e in xds_list:

        # TODO: remove this swich and use date2datenum
        if isinstance(xds_e.time.values[0], datetime):
            xds_e_dt1 = xds_e.time.values[0]
            xds_e_dt2 = xds_e.time.values[-1]
        else:
            # parse xds times to python datetime
            xds_e_dt1 = xds2datetime(xds_e.time[0])
            xds_e_dt2 = xds2datetime(xds_e.time[-1])

        if d1 == None:
            d1 = xds_e_dt1
            d2 = xds_e_dt2

        d1 = max(xds_e_dt1, d1)
        d2 = min(xds_e_dt2, d2)

    return d1, d2

def xds_further_dates(xds_list):
    '''
    returns datetime further date limits between a list of xarray.Dataset
    '''

    d1 = None
    d2 = None

    for xds_e in xds_list:

        # TODO: remove this swich and use date2datenum
        if isinstance(xds_e.time.values[0], datetime):
            xds_e_dt1 = xds_e.time.values[0]
            xds_e_dt2 = xds_e.time.values[-1]
        else:
            # parse xds times to python datetime
            xds_e_dt1 = xds2datetime(xds_e.time[0])
            xds_e_dt2 = xds2datetime(xds_e.time[-1])

        if d1 == None:
            d1 = xds_e_dt1
            d2 = xds_e_dt2

        d1 = min(xds_e_dt1, d1)
        d2 = max(xds_e_dt2, d2)

    return d1, d2

def date2yearfrac(d):
    'Returns date d in fraction of the year'

    # get timetuple obj 
    if isinstance(d, datetime):
        ttup = d.timetuple()

    elif isinstance(d, np.datetime64):
        ttup = npdt64todatetime(d).timetuple()

    elif isinstance(d, DatetimeGregorian):
        ttup = d.timetuple()

    # total number of days in year
    year_ndays = 366.0 if calendar.isleap(ttup.tm_year) else 365.0

    return ttup.tm_yday / year_ndays

def date2datenum(d):
    'Returns date d (any format) in datetime'

    # TODO: new type switch for xarray.core numpy.datetime64
    # TODO: rename to date2datetime

    if isinstance(d, datetime):
        return d

    # else get timetuple
    elif isinstance(d, np.datetime64):
        ttup = npdt64todatetime(d).timetuple()

    elif isinstance(d, DatetimeGregorian):
        ttup = d.timetuple()

    # return datetime 
    return datetime(*ttup[:6])

def get_years_months_days(time):
    '''
    Returns years, months, days of time in separete lists

    (Used to avoid problems with dates type)
    '''

    t0 = time[0]
    if isinstance(t0, (date, datetime, DatetimeGregorian)):
        ys = np.asarray([x.year for x in time])
        ms = np.asarray([x.month for x in time])
        ds = np.asarray([x.day for x in time])

    else:
        tpd = pd.DatetimeIndex(time)
        ys = tpd.year
        ms = tpd.month
        ds = tpd.day

    return ys, ms, ds


# aux. functions for teslakit hourly output

def repair_times_hourly(xds):
    'ensures that xarray.Dataset time index is rounded to nearest hour and does not repeat values'

    # round times to hour (only for np.datetime64)
    if  isinstance(xds.time.values[0], np.datetime64):
        xds['time'] = xds['time'].dt.round('H')

    # remove duplicates
    _, ix = np.unique(xds['time'], return_index=True); xds = xds.isel(time=ix)

    return xds

def hours_since(base_date, target_dates):
    'fast method for locating "target_dates" hours since "base_date"'

    return (target_dates - base_date).astype('timedelta64[h]').astype(int)

def add_max_storms_mask(xds, times_max_storms, name_mask='max_storms'):
    'fast method for adding a "max_storms" mask to a hourly xarray.Dataset'

    # find max storm indexes and make boolean mask
    ixs = hours_since(xds.time.values[0], times_max_storms)
    np_mask = np.zeros(len(xds.time.values), dtype='bool'); np_mask[ixs] = True

    # add mask to dataset
    xds[name_mask] = (('time'), np_mask)

    return xds




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from .geo import shoot


def track_from_parameters(
    pmin, vmean, delta, gamma,
    x0, y0, x1, R,
    date_ini, hours,
    great_circle=False):
    '''
    Calculates storm track variables from storm track parameters

    pmin, vmean, delta, gamma  - storm track parameters
    x0, y0                     - site coordinates (longitude, latitude)
    x1                         - enter point in computational grid
    R                          - radius (º)
    date_ini                   - initial date 'yyyy-mm-dd HH:SS'
    hours                      - number of hours to generate
    great_circle               - True for using great circle lon,lat calculation
    '''

    RE = 6378.135  # earth radius

    # generation of storm track 
    xc = x0 + R * np.sin(delta * np.pi/180)  # enter point in the smaller radius
    yc = y0 + R * np.cos(delta * np.pi/180)

    d = (x1 - xc) / np.sin(gamma * np.pi/180)
    y1 = yc + d * np.cos(gamma * np.pi/180)

    # time array for SWAN input
    time_input = pd.date_range(date_ini, periods=hours, freq='H')

    # storm track (pd.DataFrame)
    st = pd.DataFrame(index=time_input, columns=['move', 'vf', 'pn', 'p0', 'lon', 'lat'])
    st['move'] = gamma
    st['vf'] = vmean
    st['pn'] = 1013
    st['p0'] = pmin

    # calculate lon and lat
    if not great_circle:
        st['lon'] = x1 - (st['vf']*180/(RE*np.pi)) * np.sin(gamma*np.pi/180) * list(range(len(st)))
        st['lat'] = y1 - (st['vf']*180/(RE*np.pi)) * np.cos(gamma*np.pi/180) * list(range(len(st)))

    else:
        x2 = x1 - (st['vf']*180/(RE*np.pi)) * np.sin(gamma*np.pi/180) * list(range(len(st)))
        y2 = y1 - (st['vf']*180/(RE*np.pi)) * np.cos(gamma*np.pi/180) * list(range(len(st)))
        xt, yt = [], []
        for i in list(range(0,hours)):
            glon, glat, baz = shoot(x1, y1, gamma+180, vmean * i)
            xt.append(glon)
            yt.append(glat)
        st['lon'] = xt
        st['lat'] = yt

    # add some metadata
    st.x0 = x0
    st.y0 = y0
    st.R = R

    return st




from math import pi
from typing import List, Tuple, Union

import numpy as np
from shapely.geometry import Point, Polygon
from shapely.vectorized import contains

from .constants import EARTH_RADIUS_NM

# Constants
FLATTENING = 1 / 298.257223563
EPS = 0.00000000005
DEG2RAD = pi / 180.0
RAD2DEG = 180.0 / pi


# TODO: Check which functions are implemented in Pyproj library!


def convert_to_radians(*args: Union[float, np.ndarray]) -> tuple:
    """
    Convert degree inputs to radians.

    Parameters
    ----------
    *args : Union[float, np.ndarray]
        Variable number of inputs in degrees to convert to radians.
        Can be either scalar floats or numpy arrays.

    Returns
    -------
    tuple
        Tuple of input values converted to radians, preserving input types.

    Examples
    --------
    >>> convert_to_radians(90.0)
    (1.5707963267948966,)
    >>> convert_to_radians(90.0, 180.0)
    (1.5707963267948966, 3.141592653589793)
    """

    return tuple(np.radians(arg) for arg in args)


def geodesic_distance(
    lat1: Union[float, np.ndarray],
    lon1: Union[float, np.ndarray],
    lat2: Union[float, np.ndarray],
    lon2: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """
    Calculate great circle distance between two points on Earth.

    Parameters
    ----------
    lat1 : Union[float, np.ndarray]
        Latitude of first point(s) in degrees
    lon1 : Union[float, np.ndarray]
        Longitude of first point(s) in degrees
    lat2 : Union[float, np.ndarray]
        Latitude of second point(s) in degrees
    lon2 : Union[float, np.ndarray]
        Longitude of second point(s) in degrees

    Returns
    -------
    Union[float, np.ndarray]
        Great circle distance(s) in degrees

    Notes
    -----
    Uses the haversine formula to calculate great circle distance.
    The result is in degrees of arc on a sphere.

    Examples
    --------
    >>> geodesic_distance(0, 0, 0, 90)
    90.0
    >>> geodesic_distance([0, 45], [0, -90], [0, -45], [90, 90])
    array([90., 180.])
    """

    lon1, lat1, lon2, lat2 = convert_to_radians(lon1, lat1, lon2, lat2)

    a = (
        np.sin((lat2 - lat1) / 2) ** 2
        + np.cos(lat1) * np.cos(lat2) * np.sin((lon2 - lon1) / 2) ** 2
    )
    a = np.clip(a, 0, 1)

    rng = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

    return np.degrees(rng)


def geo_distance_cartesian(
    y_matrix: Union[float, np.ndarray],
    x_matrix: Union[float, np.ndarray],
    y_point: Union[float, np.ndarray],
    x_point: Union[float, np.ndarray],
) -> np.ndarray:
    """
    Returns cartesian distance between y,x matrix and y,x point.
    Optimized using vectorized operations.

    Parameters
    ----------
    y_matrix : Union[float, np.ndarray]
        2D array of y-coordinates (latitude or y in Cartesian).
    x_matrix : Union[float, np.ndarray]
        2D array of x-coordinates (longitude or x in Cartesian).
    y_point : Union[float, np.ndarray]
        y-coordinate of the point (latitude or y in Cartesian).
    x_point : Union[float, np.ndarray]
        x-coordinate of the point (longitude or x in Cartesian).

    Returns
    -------
    np.ndarray
        Array of distances in the same units as x_matrix and y_matrix.
    """

    dist = np.sqrt((y_point - y_matrix) ** 2 + (x_point - x_matrix) ** 2)

    return dist


def geodesic_azimuth(
    lat1: Union[float, np.ndarray],
    lon1: Union[float, np.ndarray],
    lat2: Union[float, np.ndarray],
    lon2: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """
    Calculate azimuth between two points on Earth.

    Parameters
    ----------
    lat1 : Union[float, np.ndarray]
        Latitude of first point(s) in degrees
    lon1 : Union[float, np.ndarray]
        Longitude of first point(s) in degrees
    lat2 : Union[float, np.ndarray]
        Latitude of second point(s) in degrees
    lon2 : Union[float, np.ndarray]
        Longitude of second point(s) in degrees

    Returns
    -------
    Union[float, np.ndarray]
        Azimuth(s) in degrees from North

    Notes
    -----
    The azimuth is the angle between true north and the direction to the second point,
    measured clockwise from north. Special cases are handled for points at the poles.

    Examples
    --------
    >>> geodesic_azimuth(0, 0, 0, 90)
    90.0
    >>> geodesic_azimuth([0, 45], [0, -90], [0, -45], [90, 90])
    array([90., 90.])
    """

    lon1, lat1, lon2, lat2 = convert_to_radians(lon1, lat1, lon2, lat2)

    az = np.arctan2(
        np.cos(lat2) * np.sin(lon2 - lon1),
        np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(lon2 - lon1),
    )

    # Handle special cases at poles
    az = np.where(lat1 <= -pi / 2, 0, az)
    az = np.where(lat2 >= pi / 2, 0, az)
    az = np.where(lat2 <= -pi / 2, pi, az)
    az = np.where(lat1 >= pi / 2, pi, az)

    return np.degrees(az % (2 * pi))


def geodesic_distance_azimuth(
    lat1: Union[float, np.ndarray],
    lon1: Union[float, np.ndarray],
    lat2: Union[float, np.ndarray],
    lon2: Union[float, np.ndarray],
) -> Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculate both great circle distance and azimuth between two points.

    Parameters
    ----------
    lat1 : Union[float, np.ndarray]
        Latitude of first point(s) in degrees
    lon1 : Union[float, np.ndarray]
        Longitude of first point(s) in degrees
    lat2 : Union[float, np.ndarray]
        Latitude of second point(s) in degrees
    lon2 : Union[float, np.ndarray]
        Longitude of second point(s) in degrees

    Returns
    -------
    Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]
        Tuple containing:
        - distance(s) : Great circle distance(s) in degrees
        - azimuth(s) : Azimuth(s) in degrees from North

    See Also
    --------
    geodesic_distance : Calculate only the great circle distance
    geodesic_azimuth : Calculate only the azimuth

    Examples
    --------
    >>> dist, az = geodesic_distance_azimuth(0, 0, 0, 90)
    >>> dist
    90.0
    >>> az
    90.0
    """

    return geodesic_distance(lat1, lon1, lat2, lon2), geodesic_azimuth(
        lat1, lon1, lat2, lon2
    )


def shoot(
    lon: Union[float, np.ndarray],
    lat: Union[float, np.ndarray],
    azimuth: Union[float, np.ndarray],
    maxdist: Union[float, np.ndarray],
) -> Tuple[
    Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]
]:
    """
    Calculate endpoint given starting point, azimuth and distance.

    Parameters
    ----------
    lon : Union[float, np.ndarray]
        Starting longitude(s) in degrees
    lat : Union[float, np.ndarray]
        Starting latitude(s) in degrees
    azimuth : Union[float, np.ndarray]
        Initial azimuth(s) in degrees
    maxdist : Union[float, np.ndarray]
        Distance(s) to travel in kilometers

    Returns
    -------
    Tuple[Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]]
        Tuple containing:
        - final_lon : Final longitude(s) in degrees
        - final_lat : Final latitude(s) in degrees
        - back_azimuth : Back azimuth(s) in degrees

    Notes
    -----
    This function implements a geodesic shooting algorithm based on
    T. Vincenty's method. It accounts for the Earth's ellipsoidal shape.

    Raises
    ------
    ValueError
        If attempting to shoot from a pole in a direction not along a meridian.

    Examples
    --------
    >>> lon_f, lat_f, baz = shoot(0, 0, 90, 111.195)  # ~1 degree at equator
    >>> round(lon_f, 6)
    1.0
    >>> round(lat_f, 6)
    0.0
    >>> round(baz, 6)
    270.0
    """

    # Convert inputs to arrays
    lon, lat, azimuth, maxdist = map(np.asarray, (lon, lat, azimuth, maxdist))

    glat1 = lat * DEG2RAD
    glon1 = lon * DEG2RAD
    s = maxdist / 1.852  # Convert km to nautical miles
    faz = azimuth * DEG2RAD

    # Check for pole condition
    pole_condition = (np.abs(np.cos(glat1)) < EPS) & ~(np.abs(np.sin(faz)) < EPS)
    if np.any(pole_condition):
        raise ValueError("Only N-S courses are meaningful, starting at a pole!")

    r = 1 - FLATTENING
    tu = r * np.tan(glat1)
    sf = np.sin(faz)
    cf = np.cos(faz)

    # Handle cf == 0 case
    b = np.zeros_like(cf)
    nonzero_cf = cf != 0
    b[nonzero_cf] = 2.0 * np.arctan2(tu[nonzero_cf], cf[nonzero_cf])

    cu = 1.0 / np.sqrt(1 + tu * tu)
    su = tu * cu
    sa = cu * sf
    c2a = 1 - sa * sa
    x = 1.0 + np.sqrt(1.0 + c2a * (1.0 / (r * r) - 1.0))
    x = (x - 2.0) / x
    c = 1.0 - x
    c = (x * x / 4.0 + 1.0) / c
    d = (0.375 * x * x - 1.0) * x
    tu = s / (r * EARTH_RADIUS_NM * c)
    y = tu.copy()

    # Iterative solution
    while True:
        sy = np.sin(y)
        cy = np.cos(y)
        cz = np.cos(b + y)
        e = 2.0 * cz * cz - 1.0
        c = y.copy()
        x = e * cy
        y = e + e - 1.0
        y = (
            ((sy * sy * 4.0 - 3.0) * y * cz * d / 6.0 + x) * d / 4.0 - cz
        ) * sy * d + tu

        if np.all(np.abs(y - c) <= EPS):
            break

    b = cu * cy * cf - su * sy
    c = r * np.sqrt(sa * sa + b * b)
    d = su * cy + cu * sy * cf
    glat2 = (np.arctan2(d, c) + pi) % (2 * pi) - pi
    c = cu * cy - su * sy * cf
    x = np.arctan2(sy * sf, c)
    c = ((-3.0 * c2a + 4.0) * FLATTENING + 4.0) * c2a * FLATTENING / 16.0
    d = ((e * cy * c + cz) * sy * c + y) * sa
    glon2 = ((glon1 + x - (1.0 - c) * d * FLATTENING + pi) % (2 * pi)) - pi
    baz = (np.arctan2(sa, b) + pi) % (2 * pi)

    return (glon2 * RAD2DEG, glat2 * RAD2DEG, baz * RAD2DEG)


def create_polygon(coordinates: List[Tuple[float, float]]) -> Polygon:
    """
    Create a polygon from a list of (longitude, latitude) coordinates.

    Parameters
    ----------
    coordinates : List[Tuple[float, float]]
        List of (longitude, latitude) coordinate pairs that define the polygon vertices.
        The first and last points should be the same to close the polygon.

    Returns
    -------
    Polygon
        A shapely Polygon object.

    Examples
    --------
    >>> coords = [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]  # Square
    >>> poly = create_polygon(coords)
    """

    return Polygon(coordinates)


def points_in_polygon(
    lon: Union[List[float], np.ndarray],
    lat: Union[List[float], np.ndarray],
    polygon: Polygon,
) -> np.ndarray:
    """
    Check which points are inside a polygon.

    Parameters
    ----------
    lon : Union[List[float], np.ndarray]
        Array or list of longitude values.
    lat : Union[List[float], np.ndarray]
        Array or list of latitude values.
        Must have the same shape as lon.
    polygon : Polygon
        A shapely Polygon object.

    Returns
    -------
    np.ndarray
        Boolean array indicating which points are inside the polygon.

    Raises
    ------
    ValueError
        If lon and lat arrays have different shapes.

    Examples
    --------
    >>> coords = [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]  # Square
    >>> poly = create_polygon(coords)
    >>> lon = [0.5, 2.0]
    >>> lat = [0.5, 2.0]
    >>> mask = points_in_polygon(lon, lat, poly)
    >>> print(mask)  # [True, False]
    """

    if isinstance(lon, list):
        lon = np.array(lon)
    if isinstance(lat, list):
        lat = np.array(lat)

    if lon.shape != lat.shape:
        raise ValueError("lon and lat arrays must have the same shape")

    # Create Point objects for each coordinate pair
    points_geom = [Point(x, y) for x, y in zip(lon, lat)]

    # Check which points are inside the polygon
    return np.array([polygon.contains(p) for p in points_geom])


def filter_points_in_polygon(
    lon: Union[List[float], np.ndarray],
    lat: Union[List[float], np.ndarray],
    polygon: Polygon,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Filter points to keep only those inside a polygon.

    Parameters
    ----------
    lon : Union[List[float], np.ndarray]
        Array or list of longitude values.
    lat : Union[List[float], np.ndarray]
        Array or list of latitude values.
        Must have the same shape as lon.
    polygon : Polygon
        A shapely Polygon object.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing:
        - filtered_lon : Array of longitudes inside the polygon
        - filtered_lat : Array of latitudes inside the polygon

    Raises
    ------
    ValueError
        If lon and lat arrays have different shapes.

    Examples
    --------
    >>> coords = [(0, 0), (0, 1), (1, 1), (1, 0), (0, 0)]  # Square
    >>> poly = create_polygon(coords)
    >>> lon = [0.5, 2.0]
    >>> lat = [0.5, 2.0]
    >>> filtered_lon, filtered_lat = filter_points_in_polygon(lon, lat, poly)
    >>> print(filtered_lon)  # [0.5]
    >>> print(filtered_lat)  # [0.5]
    """

    if isinstance(lon, list):
        lon = np.array(lon)
    if isinstance(lat, list):
        lat = np.array(lat)

    if lon.shape != lat.shape:
        raise ValueError("lon and lat arrays must have the same shape")

    mask = points_in_polygon(lon, lat, polygon)
    return lon[mask], lat[mask]


def buffer_area_for_polygon(polygon: Polygon, area_factor: float) -> Polygon:
    """
    Buffer the polygon by a factor of its area divided by its length.
    This is a heuristic to ensure that the buffer is proportional to the size of the polygon.

    Parameters
    ----------
    polygon : Polygon
        The polygon to be buffered.
    mas : float
        The buffer factor.

    Returns
    -------
    Polygon
        The buffered polygon.

    Example
    -------
    >>> from shapely.geometry import Polygon
    >>> polygon = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    >>> mas = 0.1
    >>> buffered_polygon = buffer_area_for_polygon(polygon, mas)
    >>> print(buffered_polygon)
    POLYGON ((-0.1 -0.1, 1.1 -0.1, 1.1 1.1, -0.1 1.1, -0.1 -0.1))
    """

    return polygon.buffer(area_factor * polygon.area / polygon.length)


def mask_points_outside_polygon(
    elements: np.ndarray, node_coords: np.ndarray, poly: Polygon
) -> np.ndarray:
    """
    Returns a boolean mask indicating which triangle elements have at least two vertices outside the polygon.

    This version uses matplotlib.path.Path for high-performance point-in-polygon testing.

    Parameters
    ----------
    elements : (n_elements, 3) np.ndarray
        Array containing indices of triangle vertices.
    node_coords : (n_nodes, 2) np.ndarray
        Array of node coordinates as (x, y) pairs.
    poly : shapely.geometry.Polygon
        Polygon used for containment checks.

    Returns
    -------
    mask : (n_elements,) np.ndarray
        Boolean array where True means at least two vertices of the triangle lie outside the polygon.

    Example
    -------
    >>> import numpy as np
    >>> from shapely.geometry import Polygon
    >>> elements = np.array([[0, 1, 2], [1, 2, 3]])
    >>> node_coords = np.array([[0, 0], [1, 0], [0, 1], [1, 1]])
    >>> poly = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    >>> mask = mask_points_outside_polygon(elements, node_coords, poly)
    >>> print(mask)
    [False False]
    """

    tri_coords = node_coords[elements]

    x = tri_coords[..., 0]
    y = tri_coords[..., 1]

    inside = contains(poly, x, y)

    num_inside = np.sum(inside, axis=1)

    return num_inside < 3
